import csv
import os
import pandas as pd
import statsmodels
import numpy as np
from conpagnon.utils.folders_and_files_management import create_directory
from functools import reduce
"""Text file management, output results, modify
 and append information to text files

Author: Dhaif BEKHA.
"""


def read_csv(csv_file, delimiter=','):
    """Read a CSV file and return a panda.DataFrame

     Parameters
     ----------
     csv_file: str
        The full path to the CSV file to read
     delimiter: str
        The separator use in the CSV file

    """
    # Read the csv file
    data = pd.read_csv(csv_file, delimiter=delimiter)

    return data


def csv_from_dictionary(subjects_dictionary, groupes, kinds, field_to_write,
                        header, csv_filename, output_directory, delimiter=','):
    """Write a csv file from a subjects dictionary.

    Parameters
    ----------
    subjects_dictionary: dict
        A dictionnary with the same structure as a
        subjects connectivity matrices dictionary
    groupes: list
        The list of groups to write
    kinds: list
        The list of kind to write
    field_to_write: str
        The field containing the value to write
        for each subject.
    header: list
        The header of the CSV file, in a list of
        column name
    csv_filename: str
        The end of CSV filename with the extension
    output_directory: str
        The full path to a directory for saving
        the CSV file.
    delimiter: str, optional
        The delimiter between columns.
        Default is a comma.

    """
    # Write the corresponding field for each subject
    # in each groups and kinds.
    for group in groupes:
        for kind in kinds:
            output_csv = os.path.join(
                output_directory, group + '_' + kind + '_' + csv_filename)
            with open(output_csv, 'w') as csv_file:
                # Initialize a writer object
                writer = csv.writer(csv_file, delimiter=delimiter)
                # The first row is the header
                writer.writerow(header)
                # Write for each subject, the corresponding connectivity value
                # field
                for subject, subject_sub_dictionary in subjects_dictionary[group].items():
                    writer.writerow([subject, subject_sub_dictionary[kind][field_to_write]])


def csv_from_intra_network_dictionary(subjects_dictionary, groupes, kinds, network_labels_list,
                                      field_to_write, output_directory, csv_prefix, delimiter=',',
                                      ):
    """Write csv file from the intra-network connectivity dictionary structure.

    """
    for group in groupes:
        for kind in kinds:
            for network in network_labels_list:
                # Create the output directory
                header = ['subjects', 'intra_' + network + '_connectivity']
                create_directory(directory=os.path.join(
                    output_directory, kind, network), erase_previous=False)
                # CSV filename
                csv_filename = group + '_' + csv_prefix + '_' + network + '_connectivity.csv'
                output_csv = os.path.join(
                    output_directory, kind, network, csv_filename)
                with open(output_csv, 'w') as csv_file:
                    # Initialize a writer object
                    writer = csv.writer(csv_file, delimiter=delimiter)
                    # The first row is the header
                    writer.writerow(header)
                    # Write for each subject, the corresponding connectivity value
                    # field
                    for subject, subject_sub_dictionary in subjects_dictionary[group].items():
                        writer.writerow(
                            [subject, subject_sub_dictionary[kind][network][field_to_write]])


def dataframe_to_csv(dataframe, path, delimiter=',', index=False):
    """Create and write a CSV file from a DataFrame

    """
    dataframe.to_csv(path, sep=delimiter, index=index)


def read_excel_file(excel_file_path, sheetname,
                    subjects_column_name):
    """Read a excel document

    Parameters
    ----------
    excel_file_path: str
        Full path to the excel document
    sheetname: str
        The sheetname to read in the excel document
    subjects_column_name: str
        The column name containing the subjects identifiers.

    Returns
    -------
    output: pandas.DataFrame
        A panda DataFrame, indexed by subject name.
    """

    data = pd.read_excel(io=excel_file_path, sheet_name=sheetname)
    # Shift the index
    data = data.set_index(subjects_column_name)

    return data


def shift_index_column(panda_dataframe, columns_to_index):
    """Shift the index column of a pandas DataFrame

    Parameters
    ----------
    panda_dataframe: pandas.DataFrame
        A pandas dataframe.
    columns_to_index: list
        Column label or list of column labels / arrays

    Returns
    -------
    output: pandas.DataFrame
        A new pandas DataFrame with the shifted columns as index.
    """
    new_data = panda_dataframe.set_index(columns_to_index)

    return new_data


def concatenate_dataframes(list_of_dataframes, axis=0):
    """Concatenate a list of pandas DataFrame

    """
    new_data = pd.concat(list_of_dataframes, axis=axis)

    return new_data


def merge_by_index(dataframe1, dataframe2, left_index=True, right_index=True):
    """Merge two dataframes based on the index concordances

    Parameters
    ----------
    dataframe1: pandas.DataFrame
        A panda dataframe
    dataframe2: pandas.DataFrame
        A panda dataframe
    left_index: bool, optional
        If True, the merge operation is based on the left index
    right_index: bool, optional
        If True, the merge operation is based on the right index

    Returns
    -------
    output: pandas.DataFrame
        The merged dataframe.

    Notes
    -----
    If `left_index` and `right_index` are both True
    the merge  is based on the intersection of both dataframe,
    i.e a missing index in one of the dataframe will be deleted in
    the final dataframe.
    """
    merged_dataframe = pd.merge(dataframe1, dataframe2, left_index=left_index,
                                right_index=right_index)

    return merged_dataframe


def write_ols_results(ols_fit, design_matrix, response_variable, output_dir, model_name,
                      design_matrix_index_name=None):
    """Write OLS result, along with the design matrix and the variable to explain.
    """
    # Check if we have a statmodels OLS object
    if not isinstance(ols_fit, statsmodels.regression.linear_model.RegressionResultsWrapper):
        raise TypeError('expected a statsmodels regression results wrapper '
                        'but got type {} instead'.format(type(ols_fit)))

    header = ['variables', 'coefficients', 'std_error', 't', 'p_value', 'conf_inf', 'conf_sup']
    # Construction of DataFrame based on the fit result of the model
    # we transpose it to get a column structure
    ols_results_dataframe = pd.DataFrame([ols_fit.params, ols_fit.bse, ols_fit.tvalues,
                                          ols_fit.pvalues, ols_fit.conf_int()[0],
                                          ols_fit.conf_int()[1]]).T
    ols_results_dataframe = ols_results_dataframe.reset_index()
    ols_results_dataframe.columns = header
    # We write the result in a CSV file
    with open(os.path.join(output_dir, model_name + '_parameters.csv'), 'w') as csvfile:
        # Save the results: coefficients, t-values, p-values, standard error,
        # and confidence interval
        ols_results_dataframe.to_csv(csvfile, index=False)

    with open(os.path.join(output_dir, model_name + '_qualitity_fit.csv'), 'w') as csvfile:
        csv_writer = csv.writer(csvfile)
        # Save r squared, adjusted r squared, degrees of freedom
        csv_writer.writerow(['r_squared', 'adj_r_squared', 'n_obs', 'df_model', 'df_resid'])
        csv_writer.writerow([ols_fit.rsquared, ols_fit.rsquared_adj, ols_fit.nobs, ols_fit.df_model,
                             ols_fit.df_resid])

    with open(os.path.join(output_dir, model_name + '_design_matrix.csv'), 'w') as csvfile:
        # Save the design matrix, and the response variable in the
        # same dataframe
        data = concatenate_dataframes([design_matrix, response_variable], axis=1)
        if design_matrix_index_name is not None:
            data.index.name = design_matrix_index_name

        data.to_csv(csvfile, index=True)

    # save the prediction for plotting purpose: numpy will be enough
    np.savetxt(os.path.join(output_dir, model_name + '_prediction.csv'), np.c_[ols_fit.predict()],
               header='prediction', comments='')


def group_by_factors(dataframe, list_of_factors, return_type='list_of_dataframe'):
    """Group by factors present in a dataframe

    Parameters
    ----------
    dataframe: pandas.DataFrame
        A pandas dataframe.
    list_of_factors: list
        The list of factors, i.e columns name in the dataframe,
        you want to group by.
    return_type: str
        The output format, choices are `list_of_dataframe` or
        `dictionary`. If the former, a list of dataframe is returned
        of length equal to the number of groups, if the latter a dictionary
        with groups name as keys and corresponding dataframe as values is returned.
        Default is `list_of_dataframe`.

    Returns
    -------
    output:
        A list or dictionary of the corresponding dataframe group by attribute.
    """
    # Group by the list of factors
    grouped_dataframe = dataframe.groupby(list_of_factors)
    # Get the groups keys name
    groups_names = grouped_dataframe.groups.keys()
    # Depending on the wanted return_type, construct
    # a dictionary of a list
    if return_type is 'list_of_dataframe':
        # Initialize a list containing the dataframe
        list_of_dataframes = []
        for group in groups_names:
            list_of_dataframes.append(grouped_dataframe.get_group(name=group))

        dataframe_by_group = list_of_dataframes
    elif return_type is 'dictionary':
        # Initialize a dictionary containing groups name
        # as keys, and the corresponding group dataframe as values
        grouped_dataframe_dictionary = {group_name: grouped_dataframe.get_group(name=group_name) for
                                        group_name in groups_names}

        dataframe_by_group = grouped_dataframe_dictionary
    else:
        raise ValueError('return type not understood. Choices are dictionary, list_of_dataframe'
                         'and you enter {}'.format(return_type))

    return dataframe_by_group


def dictionary_to_csv(dictionary, output_dir, output_filename):
    """Write dictionary couple (key, value) in a CSV file
    """

    with open(os.path.join(output_dir, output_filename), 'w') as f:
        w = csv.writer(f)
        w.writerows(dictionary.items())


def merge_list_dataframes(list_dataframes):
    """Merge a list of dataframes
    """
    df = reduce(lambda df1, df2: pd.merge(df1, df2, left_index=True, right_index=True),
                list_dataframes)

    return df


def remove_duplicate(seq):
    """Remove duplicate in a sequence of items while keeping
    the order.
    """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def _flatten(values):
    if isinstance(values, np.ndarray):
        yield values.flatten()
    else:
        for value in values:
            yield from _flatten(value)


def _unflatten(flat_values, prototype, offset):
    if isinstance(prototype, np.ndarray):
        shape = prototype.shape
        new_offset = offset + np.product(shape)
        value = flat_values[offset:new_offset].reshape(shape)
        return value, new_offset
    else:
        result = []
        for value in prototype:
            value, offset = _unflatten(flat_values, value, offset)
            result.append(value)
        return result, offset


def flatten(values):
    """Flatten a list of numpy ND-array

    Parameters
    ----------
    values: list
        A list of numpy array, with same or different dimensions.

    Returns
    -------
    output: numpy.array
        A flat array (one dimensional array) containing all
        the values in the same order of the list of array.

    """
    return np.concatenate(list(_flatten(values)))


def unflatten(flat_values, prototype):
    """Unflatten a one dimension array of values to the
    original list of array.

    Parameters
    ----------
    flat_values: numpy.ndarray
        The numpy array containing the values.
    prototype: list
        The original list of numpy array.

    Returns
    -------
    output: list
        A list of array with the same structure as prototype.
    """
    result, offset = _unflatten(flat_values, prototype, 0)
    assert(offset == len(flat_values))
    return result



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 19 17:02:19 2017

ComPagnon 2.0

@author: Dhaif BEKHA

"""
from conpagnon.data_handling.data_architecture import read_text_data_file, create_group_dictionnary
from conpagnon.data_handling import atlas
import numpy as np
from nilearn import input_data
from nilearn.connectome import ConnectivityMeasure, vec_to_sym_matrix
from conpagnon.utils.array_operation import masked_arrays_mean, check_2d, vectorizer
from copy import deepcopy
from conpagnon.utils import array_operation
import webcolors
import itertools
from math import sqrt
from conpagnon.data_handling.data_management import remove_duplicate
from scipy.stats import zmap, ttest_1samp
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
from conpagnon.utils.folders_and_files_management import save_object
import os
from statsmodels.stats.multitest import multipletests


def create_connectivity_mask(time_series_dictionary, groupes):
    """Create boolean mask for each subjects accounting
    for discarded roi if they exist.

    Parameters
    ----------
    time_series_dictionary: dict
        The time series dictionary structured as follow:
            - The first keys levels is the different groupes.
            - The second keys levels is the subjects IDs
            - The third levels is two keys : 'time_series' containing the
            subject time series in an array of shape (number of regions, number of time points)
            A key 'discarded_rois' containing an array of the index of ROIs
            where the corresponding labels is 'void'. If no void labels is detected,
            then the array is empty.
    groupes: list
        The list of groups in the study.

    Returns
    -------
    output: dict
        The same time series dictionary with a field called 'masked_array' for
        each subjects. A masked array is a boolean mask accounting for discarded
        rois, True for a discarded rois, and False elsewhere.

    Notes
    -----
    If 'discarded_rois' is empty the mask is False at every position
    in the array.
    Almost all the operation in ConPagnon accounts for discarded rois with
    masked array via the dedicated module in NumPy.

    """
    for group in groupes:
        subject_in_group = list(time_series_dictionary[group].keys())
        for subject in subject_in_group:
            # If a field discarded_rois exist:
            number_of_regions = time_series_dictionary[group][subject]['time_series'].shape[1]
            # mask computation
            subject_mask = np.zeros((number_of_regions, number_of_regions), dtype='bool')
            if 'discarded_rois' in time_series_dictionary[group][subject].keys():
                # We compute a mask according the position of discarded roi
                # number of regions
                empty_roi_position = time_series_dictionary[group][subject]['discarded_rois']
                # Replace by True the boolean value position corresponding to discarded rois
                # If empty_roi_position is not empty
                if empty_roi_position.size > 0:
                    for discarded_roi in empty_roi_position:
                        # True for the entire rows corresponding to the current empty roi
                        subject_mask[:, discarded_roi] = True
                        # True for the entire columns corresponding to the current empty roi
                        subject_mask[discarded_roi, :] = True
            else:
                # We compute a mask, with False everywhere, because all the test
                # is performed with masked array in general.
                subject_mask = subject_mask

            # Create a new field
            time_series_dictionary[group][subject]['masked_array'] = subject_mask

    return time_series_dictionary


def time_series_extraction_with_individual_atlases(root_fmri_data_directory, groupes, subjects_id_data_path, group_data,
                                                   repetition_time, low_pass_filtering=None, high_pass_filtering=None,
                                                   detrend_signal=True, standardize_signal=True, smooth_signal=None,
                                                   resampling_target='data', memory_level=1,
                                                   nilearn_cache_directory=None,
                                                   ):
    """Times series extractions for each subjects with an individual atlas.
    
    This function extract time series for each subjects according predefined regions in individual atlases.
    
    Parameters
    ----------
    
    root_fmri_data_directory : str
        The full path of a root directory containing one or numerous
        sub-directories where functional images are.
        
    groupes : list
        The list of groups of interest, i.e the name of the
        sub-directories containing the functional images you want 
        to study.
        
    subjects_id_data_path : str
        The full path to the data file containing the subjects IDs.
        
    group_data : dict
        A multi-levels dictionary structured as follow :
            - The first keys level is the different groups
            to study.
            - The second level keys is the subjects IDs for
            all subjects in each groupes.
            - The third level keys contain multiple field : 
            'functional_file' contain the full path to the subject
            fmri file, 'atlas_file' contain the full path to the 
            subject atlas, 'label_file' the full path to the subject
            atlas label, 'confounds_file' the full path to the subject
            confounds file if they exist, a empty list if not.
    
    repetition_time : float
        The repetition time in second, i.e the time between two volumes
        in a fmri image.
        
    low_pass_filtering : float or None, optional
        The low pass frequency in Hz cut-off for filtering the times series.
        Default is None.
    
    high_pass_filtering : float or None, optional
        The high pass frequency in Hz cut-off for filtering the time series.
        Default is None
        
    detrend_signal : bool, optional
        Detrend the time series removing the first order moment to the
        time series, i.e removing the mean signals to each time series.
        Default if True.
        
    standardize_signal : bool, optional
        Set the times series to unit variance
        Default is True.
        
    smooth_signal : float or None, optional
        The full-width half maximum in millimeters of a Gaussian
        spatial smoothing to apply to the time series.
        Default is None.
        
    resampling_target : str
        Gives the reference image which the source image image
        will be resample. 
        Choices are  : {“mask”, “maps”, “data”, None}. 
        Default is 'data'.
        
    memory_level : int, optional
        Caching parameters of functions. Default is 1.
        
    nilearn_cache_directory : str or None
        The full path which will contain the folder used to cache
        the regions extractions. If None, no cache is performing.
        Default is None.
    
    
    Returns
    -------

    output : dict
        A dictionnary structured as follow :
            - The first keys levels is the different groupes. 
            - The second keys levels is the subjects IDs
            - The third levels is two keys : 'time_series' containing the
            subject time series in an array of shape (number of regions, number of time points) 
            A key 'discarded_rois' containing an array of the index of ROIs 
            where the corresponding labels is 'void'. If no void labels is detected,
            then the array is empty.
            
    See Also
    --------
    
    data_architecture.fetch_data_with_individual_atlases : This function
    returned the organised dictionnary with all the information needed for the
    time series extraction with individual atlases. This is 
    simply the argument `group_data`.
    
    Notes
    -----
    
    The times series extraction is based on functions contain is the Nilearn
    packages. I encourage the users to consult the docstring of the following
    function for the detailed mechanism of signal extraction : nilearn.signal.clean,
    nilearn.input_data.NiftiMapsMasker.
    
    The subjects IDs file, whatever the format, should not contain any header.
    It's should have a row column of ID for each subjects.
    
    Remembers that the discarded rois are defined according to their labels
    which must be declared as 'void' in the subject atlas labels files. 
    
    References
    ----------
    
    The Nilearn official documentation on Github :
    [1] http://nilearn.github.io/index.html
    
    
    
    """
    
    # Dictionnary initialisation for each group, and subject as keys.
    times_series_dictionnary = create_group_dictionnary(subjects_id_data_path=subjects_id_data_path, groupes=groupes,
                                                        root_fmri_data_directory=root_fmri_data_directory)

    for groupe in group_data.keys():
        # For each each subject found in the group_data dictionnary, which fetch all the path for important file.
        for subject in group_data[groupe].keys():
            # We fetch the individual atlas of the subject
            subject_atlas = group_data[groupe][subject]['atlas_file'][0]
            # Fetch of discarded rois index, reading the subject atlas labels file, and looking for a 'void' label.
            subject_atlas_labels = read_text_data_file(file_path=group_data[groupe][subject]['label_file'][0])
            subject_void_rois = np.where(subject_atlas_labels[0] == 'void')[0]
            # Fetch of the subject functional file.
            subject_fmri = group_data[groupe][subject]['functional_file'][0]
            # Fetch the file containing the signal regressors, if it exist
            if group_data[groupe][subject]['counfounds_file']:
                subject_confounds = group_data[groupe][subject]['counfounds_file'][0]
            else:
                subject_confounds = None

            # Set the computation parameters for time series, via the NifTiMapsMasker class
            subject_masker = input_data.NiftiMapsMasker(
                maps_img=subject_atlas, resampling_target=resampling_target,
                smoothing_fwhm=smooth_signal, high_pass=high_pass_filtering,
                low_pass=low_pass_filtering, detrend=detrend_signal, standardize=standardize_signal,
                t_r=repetition_time, memory=nilearn_cache_directory, memory_level=memory_level)
            
            # We extract the subject time series, for each regions defined in the atlas, if confounds was found,
            # we regress them
            subject_time_series = subject_masker.fit_transform(imgs=subject_fmri, confounds=subject_confounds)
            
            # We fill the dictionnary for each and subjects, with the
            # time series array of shape (time point, number of regions),
            # and the list of index of discarded rois.
            times_series_dictionnary[groupe][subject] = {'time_series': subject_time_series,
                                                         'discarded_rois': subject_void_rois}

    times_series_dictionnary = create_connectivity_mask(time_series_dictionary=times_series_dictionnary,
                                                        groupes=groupes)

    return times_series_dictionnary


def time_series_extraction(root_fmri_data_directory, groupes, subjects_id_data_path,
                           reference_atlas, group_data, repetition_time,
                           low_pass_filtering=None, high_pass_filtering=None,
                           detrend_signal=True, standardize_signal=True,
                           smooth_signal=None, resampling_target='data', memory_level=1,
                           nilearn_cache_directory=None):
    
    """Times series extractions for each subjects on a common atlas.
    
    This function extract time series for each subjects according predefined regions on one common atlas.
    
    Parameters
    ----------
    root_fmri_data_directory : str
        The full path of a root directory containing one or numerous
        sub-directories where functional images are.
    groupes : list
        The list of groups of interest, i.e the name of the
        sub-directories containing the functional images you want 
        to study.
    subjects_id_data_path : str
        The full path to the data file containing the subjects IDs.
    group_data : dict
        A multi-levels dictionnary structured as follow :
            - The first keys level is the different groups
            to study.
            - The second level keys is the subjects IDs for
            all subjects in each groupes.
            - The third level keys contain multiple field : 
            'functional_file' contain the full path to the subject
            fmri file, 'atlas_file' contain the full path to the 
            subject atlas, 'label_file' the full path to the subject
            atlas label, 'confounds_file' the full path to the subject
            confounds file if they exist, a empty list if not.
    reference_atlas : str
        The full path to the reference atlas which will be used
        to extract signals from regions.
    repetition_time : float
        The repetition time in second, i.e the time between two volumes
        in a fmri image.
    low_pass_filtering : float or None, optional
        The low pass frequency in Hz cut-off for filtering the times series.
        Default is None.
    high_pass_filtering : float or None, optional
        The high pass frequency in Hz cut-off for filtering the time series.
        Default is None
    detrend_signal : bool, optional
        Detrend the time series removing the first order moment to the
        time series, i.e removing the mean signals to each time series.
        Default if True.
    standardize_signal : bool, optional
        Set the times series to unit variance
        Default is True.
    smooth_signal : float or None, optional
        The full-width half maximum in millimeters of a Gaussian
        spatial smoothing to apply to the time series.
        Default is None.
    resampling_target : str
        Gives the reference image which the source image image
        will be resample. 
        Choices are  : {“mask”, “maps”, “data”, None}. 
        Default is 'data'.
    memory_level : int, optional
        Caching parameters of functions. Default is 1.
    nilearn_cache_directory : str or None
        The full path which will contain the folder used to cache
        the regions extractions. If None, no cache is performing.
        Default is None.

    Returns
    -------
    output : dict
        A dictionnary structured as follow :
            - The first keys levels is the different groupes. 
            - The second keys levels is the subjects IDs
            - The third levels is two keys : 'time_series' containing the
            subject time series in an array of shape (number of regions, number of time points) 

    See Also
    --------
    data_architecture.fetch_data : This function
    returned the organised dictionnary with all the information needed for the
    time series extraction on a common atlas. This is 
    simply the argument `group_data`.
    
    Notes
    -----
    The times series extraction is based on functions contain is the Nilearn
    packages. I encourage the users to consult the docstring of the following
    function for the detailed mechanism of signal extraction : nilearn.signal.clean,
    nilearn.input_data.NiftiMapsMasker.
    
    The subjects IDs file, whatever the format, should not contain any header.
    It's should have a row column of ID for each subjects.
        
    References
    ----------
    The Nilearn official documentation on Github :
    [1] http://nilearn.github.io/index.html
    """

    # Dictionnary initialisation for time series, for each group, and for each subjects.
    times_series_dictionnary = create_group_dictionnary(subjects_id_data_path=subjects_id_data_path,
                                                        groupes=groupes,
                                                        root_fmri_data_directory=root_fmri_data_directory)

    for groupe in group_data.keys():
        # For each subject in the group_data dictionnary
        for subject in group_data[groupe].keys():
            # Fetch the subject functional file path
            subject_fmri = group_data[groupe][subject]['functional_file'][0]
            # Fetch the subject regressors file path, if it exist
            if group_data[groupe][subject]['counfounds_file']:
                subject_confounds = group_data[groupe][subject]['counfounds_file'][0]
            else:
                subject_confounds = None
            
            # Set the computation parameters for time series, via the NifTiMapsMasker class
            subject_masker = input_data.NiftiMapsMasker(maps_img=reference_atlas, resampling_target=resampling_target,
                                                        smoothing_fwhm=smooth_signal, high_pass=high_pass_filtering,
                                                        low_pass=low_pass_filtering, detrend=detrend_signal,
                                                        standardize=standardize_signal, t_r=repetition_time,
                                                        memory=nilearn_cache_directory, memory_level=memory_level)
            
            # We extract the subject time series, for each regions defined in the atlas, if confounds was found,
            # we regress them
            subject_time_series = subject_masker.fit_transform(imgs=subject_fmri, confounds=subject_confounds)
            
            # We fill the time series array of shape (times points, number of regions) for each subject in each group in
            # time series dictionnary
            times_series_dictionnary[groupe][subject] = {'time_series': subject_time_series}

    times_series_dictionnary = create_connectivity_mask(time_series_dictionary=times_series_dictionnary,
                                                        groupes=groupes)

    return times_series_dictionnary


def individual_connectivity_matrices(time_series_dictionary, kinds, covariance_estimator, vectorize=False,
                                     discarding_diagonal=False, z_fisher_transform=False):
    """Compute the connectivity matrices for groups of subjects
    
    This function computes connectivity matrices for different metrics.

    Parameters
    ----------
    time_series_dictionary : dict
        A multi-levels dictionnary organised as follow :
            - The first keys levels is the different groupes in the study. 
            - The second keys levels is the subjects IDs
            - The third levels is two keys : 'time_series' containing the
            subject time series in an array of shape (number of regions, number of time points) 
            A key 'discarded_rois' containing an array of the index of ROIs 
            where the corresponding labels is 'void'. If no void labels is detected,
            then the array should be empty.
    kinds : list
        List of the different metrics you want to compute the 
        connectivity matrices. Choices are 'tangent', 'correlation',
        'partial correlation', 'covariance', 'precision'.
    covariance_estimator : estimator object
        All the kinds  are based on derivation of covariances matrices. You need
        to precise the estimator, see Notes.
    vectorize : bool, optional
        If True, the connectivity matrices are reshape into 1D arrays of 
        the vectorized lower part of the matrices. Useful for classification,
        regression...
        Default is False.
    z_fisher_transform: bool, optional
        If True, the z fisher transform is apply to all
        the connectivity matrices. Default is False
    discarding_diagonal: bool, optional
        If True, the diagonal is discarded in
        the vectorization process. Default is False.

    Returns
    -------
    output : dict
        A multi-levels dictionnary organised as follow :
            - The first keys levels is the different groupes in the study. 
            - The second keys levels is the subjects IDs
            - The third levels contain multiple keys : multiple
            kind keys containing the corresponding kind matrix.
            a 'discarded_rois' key containing the index of discarded
            rois.
            Finally you should find a 'masked_array' key : This key 
            contains a array of Boolean, shape (numbers of regions, numbers of regions)
            where the value are True for the index in 'discarded_rois' array, and False
            elsewhere. See Note for further details. The diagonal of the
            matrix are saved for each kind too in a dedicated field.
            
    See Also
    --------
    time_series_extraction, time_series_extraction_with_individual_atlases :
        These are the functions which extract the time series according atlas
        regions and returned a structured dictionnary. 
        This is simply the argument `time_series_dictionary`.
        
    Notes 
    -----
    Covariances estimator are estimators compute in the scikit-learn library.
    Multiple estimator can be found in the module sklearn.covariance, popular
    choices are the Ledoit-Wolf estimator, or the OAS estimator.
    In the output dictionnary, each subjects have a masked array of boolean. The
    masked will be useful when computing the mean connectivity matrices, we will
    account the discarded rois in the derivation. For the statistical test you might
    want perform, it will be useful too, to discarded those rois. A True value is
    a masked roi, False elsewhere.
    For the tangent kind, the derivation of individual matrices need to be
    made on the POOLED GROUPS which is performed here.
    The derivation of connectivity matrices are based on Nilearn functions. I
    encourage the user to read the following docstring of important functions :
    nilearn.connectome.ConnectivityMeasure
    
    References
    ----------
    For the use of tangent :
        .. [1] G. Varoquaux et al. “Detection of brain functional-connectivity
        difference in post-stroke patients using group-level covariance modeling", MICCAI 2010
    """

    saving_subjects_connectivity_matrices_dictionary = dict.fromkeys(list(time_series_dictionary.keys()))

    for groupe in time_series_dictionary.keys():
        subjects_list = time_series_dictionary[groupe].keys()
        saving_subjects_connectivity_matrices_dictionary[groupe] = dict.fromkeys(subjects_list)
        for subject in subjects_list:
            subject_time_series = [time_series_dictionary[groupe][subject]['time_series']]
            saving_subjects_connectivity_matrices_dictionary[groupe][subject] = dict.fromkeys(kinds)

            # operation on the boolean mask for each subject
            subject_mask = time_series_dictionary[groupe][subject]['masked_array']
            if vectorize:
                if discarding_diagonal:
                    subject_mask_diagonal_, subject_mask_ = \
                        vectorizer(numpy_array=time_series_dictionary[groupe][subject]['masked_array'],
                                   discard_diagonal=discarding_diagonal,
                                   array_type='boolean')
                    saving_subjects_connectivity_matrices_dictionary[groupe][subject]['masked_array'] = \
                        subject_mask_
                    saving_subjects_connectivity_matrices_dictionary[groupe][subject][
                        'diagonal_mask'] = subject_mask_diagonal_
                else:
                    # discarding_diagonal is False
                    subject_mask_diagonal_, subject_mask_ = \
                        vectorizer(numpy_array=subject_mask,
                                   discard_diagonal=discarding_diagonal,
                                   array_type='boolean')
                    saving_subjects_connectivity_matrices_dictionary[groupe][subject]['masked_array'] = \
                        subject_mask_
                    saving_subjects_connectivity_matrices_dictionary[groupe][subject][
                        'diagonal_mask'] = subject_mask_diagonal_
            else:
                saving_subjects_connectivity_matrices_dictionary[groupe][subject]['masked_array'] = subject_mask
                saving_subjects_connectivity_matrices_dictionary[groupe][subject]['diagonal_mask'] = \
                    np.diag(subject_mask)

            # Fetch the index of discarded rois
            if 'discarded_rois' in time_series_dictionary[groupe][subject].keys():
                saving_subjects_connectivity_matrices_dictionary[groupe][subject]['discarded_rois'] = \
                    time_series_dictionary[groupe][subject]['discarded_rois']

            for kind in kinds:
                # We compute connectivity matrices except for the tangent kind, which need the pooled groups
                if kind != 'tangent':
                    kind_measure = ConnectivityMeasure(kind=kind, cov_estimator=covariance_estimator)
                    # the current subject connectivity matrices
                    subject_kind_matrice = kind_measure.fit_transform(subject_time_series)
                    # We fill the subject dictionnary, accounting for vectorization option
                    if vectorize:
                        # if discard_diagonal is True
                        if discarding_diagonal:
                            subject_kind_matrice_diagonal_, subject_kind_matrice_ = \
                                vectorizer(numpy_array=subject_kind_matrice[0, ...],
                                           discard_diagonal=discarding_diagonal,
                                           array_type='numeric')
                        else:
                            subject_kind_matrice_diagonal_, subject_kind_matrice_ = \
                                vectorizer(numpy_array=subject_kind_matrice[0, ...], discard_diagonal=False,
                                           array_type='numeric')
                        if z_fisher_transform is True:
                            # Apply a fisher transform, i.e arctanh to each connectivity coefficient
                            undefined_value = np.invert(subject_kind_matrice_ == 1)
                            subject_kind_matrice_ = np.arctanh(subject_kind_matrice_, where=undefined_value)
                            # Fill with nan, the undefined value after transformation
                            subject_kind_matrice_[np.invert(undefined_value)] = np.nan
                        else:
                            subject_kind_matrice_ = subject_kind_matrice_

                        saving_subjects_connectivity_matrices_dictionary[groupe][subject][kind] = \
                            subject_kind_matrice_[:]
                        # Save the diagonal in this case:
                        saving_subjects_connectivity_matrices_dictionary[groupe][subject][kind + '_diagonal'] = \
                            subject_kind_matrice_diagonal_
                    else:
                        # If we do not vectorize the connectivity matrix
                        if z_fisher_transform is True:
                            # Apply a fisher transform, i.e arctanh to each connectivity coefficient
                            # For r = 1, in the diagonal, arctanh is not defined.
                            # We compute a mask where False value are ignored
                            undefined_value = np.invert(subject_kind_matrice[0, ...] == 1)
                            subject_kind_matrice = np.arctanh(subject_kind_matrice, where=undefined_value)
                            subject_kind_matrice[0, np.invert(undefined_value)] = np.nan
                        else:
                            subject_kind_matrice = subject_kind_matrice

                        saving_subjects_connectivity_matrices_dictionary[groupe][subject][kind] = \
                            subject_kind_matrice[0, ...]

    if 'tangent' in kinds:
        # We call pooled_groups_connectivity on the tangent space, returning along the tangent matrices
        # the labels subjects in the same order of tangent connectivity matrices:
        tangent_matrix, subjects_order = pooled_groups_connectivity(
            time_series_dictionary=time_series_dictionary,
            kinds=['tangent'],
            covariance_estimator=covariance_estimator,
            vectorize=False)

        for groupe in time_series_dictionary.keys():
            group_subject_list = time_series_dictionary[groupe].keys()
            for subject in group_subject_list:
                subjects_order_list = list(subjects_order)
                sub_index = subjects_order_list.index(subject)
                if vectorize:
                    if discarding_diagonal:
                        diagonal_matrice_tangent, corresponding_tangent_matrice = \
                            vectorizer(numpy_array=tangent_matrix['tangent'][sub_index, ...],
                                       discard_diagonal=discarding_diagonal,
                                       array_type='numeric')
                    else:
                        diagonal_matrice_tangent, corresponding_tangent_matrice = \
                            vectorizer(numpy_array=tangent_matrix['tangent'][sub_index, ...],
                                       discard_diagonal=False,
                                       array_type='numeric')
                    saving_subjects_connectivity_matrices_dictionary[groupe][subject]['tangent'] = \
                        corresponding_tangent_matrice
                    saving_subjects_connectivity_matrices_dictionary[groupe][subject]['tangent' + '_diagonal'] = \
                        diagonal_matrice_tangent
                else:
                    corresponding_tangent_matrice = tangent_matrix['tangent'][sub_index, ...]
                    saving_subjects_connectivity_matrices_dictionary[groupe][subject]['tangent'] = \
                        corresponding_tangent_matrice

    return saving_subjects_connectivity_matrices_dictionary


def pooled_groups_tangent_mean(time_series_dictionary, covariance_estimator):
    """Compute the geometric mean of covariances connectivity matrices for the tangent kind.
    
    The geometric mean is the point in the symmetric manifold matrices spaces 
    where the tangent space is defined. Therefore it's make sense for the pooled
    groups only.
    
    Parameters
    ----------
    time_series_dictionary : dict
        A multi-levels dictionnary organised as follow :
            - The first keys levels is the different groupes in the study. 
            - The second keys levels is the subjects IDs
            - The third levels is two keys : 'time_series' containing the
            subject time series in an array of shape (number of regions, number of time points) 
            A key 'discarded_rois' containing an array of the index of ROIs 
            where the corresponding labels is 'void'. If no void labels is detected,
            then the array should be empty.
    covariance_estimator : estimator object
    All the kinds  are based on derivation of covariances matrices. You need
    to precise the estimator, see Notes.
    
    Returns 
    -------
    output : numpy.array of shape (n_features, n_features)
        The geometric mean a the pooled group.
        
    See Also
    --------
    time_series_extraction, time_series_extraction_with_individual_atlases :
    These are the functions which extract the time series according atlas
    regions and returned a structured dictionnary. This is simply the argument
    `time_series_dictionary`.
    
    Notes
    -----
    Covariances estimator are estimators compute in the scikit-learn library.
    Multiple estimator can be found in the module sklearn.covariance, popular
    choices are the Ledoit-Wolf estimator, or the OAS estimator.
    For now, we doesnt account for discarded rois for the derivation of the
    geometric mean.
    """

    tangent_measure = ConnectivityMeasure(kind='tangent', cov_estimator=covariance_estimator)
    group_stacked_time_series = []

    for groupe in time_series_dictionary.keys():
        for subject in time_series_dictionary[groupe].keys():
            # appending subject time series
            group_stacked_time_series.append(time_series_dictionary[groupe][subject]['time_series'])

    # Projecting in tangent space the whole cohort of subjects at the geometric mean
    _ = tangent_measure.fit_transform(group_stacked_time_series)
    # Retrieve the geometric mean
    tangent_mean_ = tangent_measure.mean_

    return tangent_mean_


def group_mean_connectivity(subjects_connectivity_matrices, kinds, axis=0):
    """Compute the mean connectivity matrices for each kind accounting for masked rois.
        Read the notes for the tangent space !
    
    Parameters
    ----------
    subjects_connectivity_matrices : dict
         A multi-levels dictionnary organised as follow :
             - The first keys levels is the different groupes in the study. 
            - The second keys levels is the subjects IDs
            - The third levels is the different kind matrices
            for each subjects, a 'discarded_rois' key for the
            discarded rois array index, a 'masked_array' key containing
            the array of Boolean of True for the discarded_rois index, and False
            elsewhere.
    kinds : list
        List of kinds you want the mean connectivity. Choices
        are 'correlation', 'tangent', 'covariances', 'precision', 'partial 
        correlation'. Off course, the kind should be in the subjects_connectivity_matrices
        dictionnary.
    axis : int, optional
        The axis you want to compute the mean, the subjects axis. Default is 0.
    
    Returns 
    -------
    output : dict
        A multi-levels dictionnary organised as follow :
            - The first keys levels is the different groups in the study.
            - The second keys levels is the mean connectivity matrices for
            the different kinds. They are array of shape (number of regions , number of regions)
            if vectorize is False, and shape (n_columns * (n_columns + 1) /2) else.
            
    See Also
    --------
    individual_connectivity_matrices : These function returned a organised
    dictionnary containing the connectivity matrices for different kinds. This is
    simply the argument `subjects_connectivity_matrices`
    
    Notes
    -----
    When computing the mean, we account for the 'discarded_rois' entries. That mean 
    when the value is True in the masked_array, we discard the rois for the corresponding
    subject in the derivation of the mean.
    When I compute the mean in the tangent space, it's a arithmetic mean. This
    mean matrix is in the tangent space, that is NOT in the same space as correlation or partial correlation matrix.
    Be careful with the interpretation !!
    That said, the tangent space is defined at ONE point in the manifold of symmetric matrices.
    This point is the geometric mean for the POOLED groups if multiple group are studied !
    """

    group_mean_connectivity_matrices = dict.fromkeys(list(subjects_connectivity_matrices.keys()))
    for groupe in subjects_connectivity_matrices.keys():
        # Dictionnary initialisation : for each group, we generate the list of kinds as keys
        group_mean_connectivity_matrices[groupe] = dict.fromkeys(kinds)
        # Subject list of the current group
        subjects_list = subjects_connectivity_matrices[groupe].keys()
        for kind in kinds:
            # We stack the subjects connectivity matrices for the group and current kind
            kind_stack_matrices = np.array([subjects_connectivity_matrices[groupe][subject][kind]
                                            for subject in subjects_list])
            # We also stack the subjects boolean array
            mask_array_stack = np.array([subjects_connectivity_matrices[groupe][subject]['masked_array']
                                         for subject in subjects_list])
            # We compute the mean for each kind, accounting for the discarded rois,
            # via the function call masked_array_mean
            group_mean_connectivity_matrices[groupe][kind] = masked_arrays_mean(arrays=kind_stack_matrices,
                                                                                masked_array=mask_array_stack,
                                                                                axis=axis)

    return group_mean_connectivity_matrices


def pooled_groups_connectivity(time_series_dictionary, kinds, covariance_estimator, vectorize):
    """Compute connectivity matrices of pooled groups.
    
    This function simply stack the all the times series of each group in one pooled groups and compute 
    the connectivity matrices  on the whole group. When computing the tangent kind this function is
    called.
    
    Parameters 
    ----------
    time_series_dictionary : dict
    A multi-levels dictionnary organised as follow :
        - The first keys levels is the different groupes in the study. 
        - The second keys levels is the subjects IDs
        - The third levels is two keys : 'time_series' containing the
        subject time series in an array of shape (number of regions, number of time points) 
        A key 'discarded_rois' containing an array of the index of ROIs 
        where the corresponding labels is 'void'. If no void labels is detected,
        then the array should be empty.
    kinds : list
        List of the different metrics you want to compute the 
        connectivity matrices. Choices are 'tangent', 'correlation',
        'partial correlation', 'covariance', 'precision'.
    covariance_estimator : estimator object
        All the kinds  are based on derivation of covariances matrices. You need
        to precise the estimator, see Notes.
    vectorize : bool
        If True, the connectivity matrices are reshape into 1D arrays of 
        the vectorized lower part of the matrices. Useful for classification,
        regression...
        Diagonal are discarded.
    
    Returns
    -------
    output 1 : dict
        A multi-levels dictionnary organised as follow :
            - The first level keys is the different kinds
            - The second levels is simply a ndimensional array of connectivity
            matrices of shape (number of subjects, number of regions, number of regions)
            if vectorize is False, and shape (n_columns * (n_columns + 1) /2) else
    output 2 : list
        The list of subject IDs, in the order of time series computation.
    
    Notes
    -----
    Covariances estimator are estimators compute in the scikit-learn library.
    Multiple estimator can be found in the module sklearn.covariance, popular
    choices are the Ledoit-Wolf estimator, or the OAS estimator.
    """
    group_stacked_time_series = []
    for groupe in time_series_dictionary.keys():
        subject_list = time_series_dictionary[groupe].keys()
        for subject in subject_list:
            # Appending a tuple: time series and the corresponding labels of the subjects
            group_stacked_time_series.append((time_series_dictionary[groupe][subject]['time_series'], subject))
    
    # We compute the connectivity matrices for each subject in the pooled group
    pooled_connectivity_matrices_dictionary = dict.fromkeys(kinds)
    stacked_couple_subject_time_series = np.array(group_stacked_time_series)
    group_stacked_time_series = stacked_couple_subject_time_series[:, 0]
    subject_labels = stacked_couple_subject_time_series[:, 1]
    for kind in kinds:
        connectivity_measure = ConnectivityMeasure(kind=kind,
                                                   cov_estimator=covariance_estimator,
                                                   vectorize=vectorize)
        pooled_connectivity_matrices_dictionary[kind] = \
            connectivity_measure.fit_transform(group_stacked_time_series)
        
    return pooled_connectivity_matrices_dictionary, subject_labels


def extract_sub_connectivity_matrices(subjects_connectivity_matrices, kinds,
                                      regions_index,
                                      vectorize=False,
                                      discard_diagonal=False):
    """Extract sub matrices given region index.

    Parameters
    ----------
    subjects_connectivity_matrices: dict
        The dictionnary containing for each subjects, for each group and kind, the
        connectivity matrices, of shape (n_features, n_features), and also
        the boolean mask array indicating the discarded roi.
    kinds: list
        The list of kinds.
    regions_index: list, or 1D numpy.array of shape (number of indices, )
        The list of index of regions you want to extract the connectivity. The
        region index also correspond to the index of ROIs in the 4D reference
        atlas of the study.
    vectorize: bool, optional
        If True, the extracted sub-matrices are vectorized, keeping
        the diagonal of the matrix, and the corresponding boolean
        mask.
    discard_diagonal: bool, optional
        If True, the diagonal is discard when the extracted connectivity matrices are vectorized.

    Returns
    -------
    output: dict
        A dictionnary containing for each group and kinds, for each subject
        the extract sub-matrice, and the corresponding boolean mask.
    """

    # Copy of the subjects connectivity matrices dictionnary
    subjects_connectivity_matrices_copy = deepcopy(subjects_connectivity_matrices)

    for groupe in subjects_connectivity_matrices_copy.keys():
        for subject in subjects_connectivity_matrices_copy[groupe].keys():
            for kind in kinds:
                # Sub extraction of connectivity matrices according the provided region_index array
                row_sliced = subjects_connectivity_matrices_copy[groupe][subject][kind][regions_index, :]
                subjects_connectivity_matrices_copy[groupe][subject][kind] = row_sliced[:, regions_index]
                if vectorize:
                    _, subjects_connectivity_matrices_copy[groupe][subject][kind] = \
                        vectorizer(numpy_array=row_sliced[:, regions_index],
                                   discard_diagonal=discard_diagonal,
                                   array_type='numeric')

            # Sub extraction of corresponding masked array
            row_mask_sliced = subjects_connectivity_matrices_copy[groupe][subject]['masked_array'][regions_index, :]
            subjects_connectivity_matrices_copy[groupe][subject]['masked_array'] = row_mask_sliced[:, regions_index]
            if vectorize:
                _, subjects_connectivity_matrices_copy[groupe][subject]['masked_array'] = vectorizer(
                    numpy_array=row_mask_sliced[:, regions_index],
                    discard_diagonal=discard_diagonal,
                    array_type='boolean')

    # Return the subjects sub connectivity matrices
    subjects_sub_connectivity_matrices = subjects_connectivity_matrices_copy

    return subjects_sub_connectivity_matrices


def subjects_mean_connectivity_(subjects_individual_matrices_dictionnary,
                                connectivity_coefficient_position, kinds,
                                groupes):
    """Compute for each subjects, the mean connectivity for some connectivity coefficient in the general subjects
    connectivity matrices.

    Parameters
    ----------
    subjects_individual_matrices_dictionnary: dict
        The subjects connectivity dictionnary containing connectivity matrices
        and corresponding mask array for discarded rois, for each group.
    connectivity_coefficient_position: numpy.array of shape (number of rois, row_index, column_index)
        The array containing the position in the connectivity matrices of the rois you want
        to extract the connectivity coefficients.
    kinds: list
        The list of kinds.
    groupes: list
        The list of the two group in the study.

    Returns
    -------
    output: dict
        A dictionnary containing for each subject in group, a masked array structure containing
        the numerical array of the connectivity coefficient of interest, along with the boolean
        mask accounting for discarded rois. The second key, contain the mean of the extracted
        connectivity coefficient for each subject, also accounting for discarded rois it they
        exist.

    Notes
    -----
    The subjects connectivity matrices shouldn't be vectorized, the shape should
    be (n_features, n_features).
    """
    
    # Indices of the connectivity coefficient in terms of row and column index.
    row_index = connectivity_coefficient_position[:, 0]
    column_index = connectivity_coefficient_position[:, 1]

    # Extract from the subject connectivity dictionnary the connectivity coefficients

    # Dictionnary initialisation to store the regions connectivity coefficients, and the mask value for further
    # accounting for discarded rois
    connectivity_of_interest = dict.fromkeys(list(subjects_individual_matrices_dictionnary.keys()))
    for groupe in groupes:
        subjects_list = subjects_individual_matrices_dictionnary[groupe].keys()
        connectivity_of_interest[groupe] = dict.fromkeys(list(subjects_list))
        for subject in subjects_list:
            connectivity_of_interest[groupe][subject] = dict.fromkeys(kinds)
            for kind in kinds:
                # Check is subject matrice is two dimensional
                check_2d(numpy_array=subjects_individual_matrices_dictionnary[groupe][subject][kind])
                # Check is the corresponding mask is two dimensional
                check_2d(numpy_array=subjects_individual_matrices_dictionnary[groupe][subject]['masked_array'])

                # Extract connectivity coefficient based on the list of indices of row and column index
                subjects_connectivity_of_interest = \
                    subjects_individual_matrices_dictionnary[groupe][subject][kind][row_index,
                                                                                    column_index]
                # Extract corresponding mask values
                subjects_masked_array_connectivity_of_interest = \
                    subjects_individual_matrices_dictionnary[groupe][subject]['masked_array'][row_index,
                                                                                              column_index]
                # Build a masked array structure for further computation accounting for discarded roi
                subjects_connectivity_of_interest_ma = np.ma.array(data=subjects_connectivity_of_interest,
                                                                   mask=subjects_masked_array_connectivity_of_interest)
                # Compute the mean connectivity for the current subject, accounting for discarded rois
                mean_subject_connectivity_of_interest = subjects_connectivity_of_interest_ma.mean()
                # Fill the dictionnary, saving the subject masked array structure and
                # mean connectivity of extracted coefficient.
                connectivity_of_interest[groupe][subject][kind] = \
                    {'connectivity masked array': subjects_connectivity_of_interest_ma,
                     'mean connectivity': mean_subject_connectivity_of_interest}

    return connectivity_of_interest


def intra_network_functional_connectivity(subjects_individual_matrices_dictionnary,
                                          groupes, kinds, atlas_file,
                                          network_column_name,
                                          sheetname,
                                          roi_indices_column_name, color_of_network_column):
    """Compute for each subjects, the intra network connectivity for each network in the study.

    Parameters
    ----------
    subjects_individual_matrices_dictionnary: dict
        The subjects connectivity dictionnary containing connectivity matrices
        and corresponding mask array for discarded rois, for each group.
    groupes: list
        The list of groups in the study
    kinds: list
        The list of kinds in the study
    atlas_file: string
        The full path to an excel file containing information on the atlas
    network_column_name: string
        The name of the column in the excel file containing the network label for each roi
    roi_indices_column_name: string
        The name of the column in the excel file containing the index of each ROI in the 4D atlas file.
    sheetname: string
        The name of the active sheet in the excel file
    color_of_network_column: string
        The name of the columns containing for each roi, the corresponding network color.

    Returns
    -------
    output 1: dict
        A dictionnary structure containing for each subject and for each network: the network connectivity,
        the vectorized array of coefficients of the network without the diagonal, the corresponding vectorized
        mask array accounting for discarded rois, the diagonal of the mask array, the masked array structure
        of the network, and finally the number of rois in the network.
    output 2: dict
        A dictionnary network, containing information fetch from the atlas excel file of the atlas.
    output 3: list
        The network label list.
    output 4: array of shape (number of network, 3)
        The array containing the color in the normalized RGB space, in the order of the network label list.

    Notes
    -----
    The intra connectivity is simply defined as the mean, for each network,
    of the coefficient belonging to the network. Because connectivity metrics are symmetric,
    we only taking the vectorize part of the network connectivity matrices. We account for discarded roi,
    as we compute the mean on a numpy masked array structure, that is the vectorized array
    along with the vectorized boolean mask for the current network.

    References
    ----------
    This intra-network composite scores is used in the following references:

    .. [1] M. Brier, "Loss of Intranetwork and Internetwork Resting State Functional Connections with Alzheimer's
       Disease Progression" The Journal of Neuroscience, 2012.
    .. [2] P. Wang, "Aberrant intra- and inter-network connectivity architectures in Alzheimer's disease and mild
       cognitive impairment", Nature Publishing Group, 2015.

    """

    network_dict = atlas.fetch_atlas_functional_network(atlas_excel_file=atlas_file,
                                                        sheetname=sheetname,
                                                        network_column_name=network_column_name)
    network_labels_list = list(network_dict.keys())
    # The list of network are NOT in the same order of the atlas in the analysis, 
    # we have to fetch the color according to network labels list.
    
    network_label_color_name = [list(set(network_dict[n]['dataframe'][color_of_network_column]))[0]
                                for n in network_labels_list]
    network_label_colors = (1/255)*np.array([webcolors.name_to_rgb(network_label_color_name[i])
                                             for i in range(len(network_label_color_name))])
    
    # For each subject in each group : the network intra-connectivity
    # is defined as the mean of the connectivity coefficient inside
    # the network.

    # Intra network dictionnary initialisation
    intra_network_connectivity_dict = dict.fromkeys(groupes)
    for groupe in groupes:
        subjects_list = subjects_individual_matrices_dictionnary[groupe].keys()
        intra_network_connectivity_dict[groupe] = dict.fromkeys(subjects_list)
        for subject in subjects_list:
            intra_network_connectivity_dict[groupe][subject] = dict.fromkeys(kinds)
            for kind in kinds:
                intra_network_connectivity_dict[groupe][subject][kind] = dict.fromkeys(network_labels_list)

    # Initialisation of the list containing the label of network which have at least two regions
    valid_network_list = []
    # Initialisation of the list containing the color of the label network which have at least two regions
    valid_network_color_list = []
    for network in network_labels_list:

        # Fetch the indices of the network roi in the 4D atlas file, it will the same indices to slice in the subject
        # connectivity matrices
        network_roi_4d_indices = network_dict[network]['dataframe'][roi_indices_column_name]
        # Fetch the number of rois in the network
        n_roi_in_network = network_dict[network]['number of rois']
        # Computation make sense only they are at least two regions in the current network
        if n_roi_in_network >= 2:
            # It is a valid network, we append it's label and color for plotting purpose in the t-test
            valid_network_list.append(network)
            valid_network_color_list.append(network_label_colors[network_labels_list.index(network)])
            # Extract the sub matrix corresponding to the current network for all kinds
            network_connectivity_matrix = extract_sub_connectivity_matrices(
                subjects_connectivity_matrices=subjects_individual_matrices_dictionnary,
                kinds=kinds, regions_index=network_roi_4d_indices, vectorize=False)
            # For Each group, each kind and subject:
            # sum the lower diagonal of each network, accounting for the discarded ROIs,
            # if they are some in the network
            for groupe in groupes:
                subjects_list = network_connectivity_matrix[groupe].keys()
                for subject in subjects_list:
                    for kind in kinds:
                        # Vectorize the network connectivity matrix
                        diag_network, vectorize_network = array_operation.vectorizer(
                            numpy_array=network_connectivity_matrix[groupe][subject][kind],
                            discard_diagonal=True,
                            array_type='numeric')
                        # Vectorize the boolean mask
                        diag_network_mask, vectorize_network_mask = array_operation.vectorizer(
                            numpy_array=network_connectivity_matrix[groupe][subject]['masked_array'],
                            discard_diagonal=True,
                            array_type='boolean')
                        # Construct the masked array structure for the current network
                        network_masked_array = np.ma.array(data=vectorize_network, mask=vectorize_network_mask)
                        # Compute the mean for the network, accounting for discarded
                        # rois via the numpy masked array structure
                        network_mean = network_masked_array.mean()
                        # Fill the intra network dictionnary
                        intra_network_connectivity_dict[groupe][subject][kind][network] = {
                            'network connectivity strength': network_mean,
                            'network array': vectorize_network,
                            'network diagonal array': diag_network,
                            'network mask': vectorize_network_mask,
                            'network diagonal masked array': diag_network_mask,
                            'number of rois in network': n_roi_in_network}

    return intra_network_connectivity_dict, network_dict, valid_network_list,  valid_network_color_list


def inter_network_subjects_connectivity_matrices(subjects_individual_matrices_dictionnary, groupes,
                                                 kinds, atlas_file, sheetname,
                                                 network_column_name, roi_indices_column_name):
    """Compute for each subjects, the inter network connectivity matrices

    Parameters
    ----------
    subjects_individual_matrices_dictionnary: dict
        The subjects connectivity matrices dictionnary for each groupes, and kind in the study.
    groupes: list
        The list of groups in the study
    kinds: list
        The list of kinds in the study
    atlas_file: string
        The full path to an excel file containing information on the atlas
    network_column_name: string
        The name of the column in the excel file containing the network label for each roi
    roi_indices_column_name
        The name of the column in the excel file containing the index of each ROI in the 4D atlas file.
    sheetname: string
        The name of the active sheet in the excel file

    Returns
    -------
    output: dict
        The subjects inter network connectivity matrices dictionnary, for each group and kinds. Each matrices
        should have shape (number of network, number of network).

    Notes
    -----
    The inter-network connectivity is simply defined as the mean of all possible
    connection between a network pair. We also account for the possible discarded rois
    in one or both the network by computing the mean on a masked array structure, that is,
    the inter-network coefficient of interest, along with the corresponding value of the masked
    array.

    References
    ----------
    This inter-network composite score is used in the following references :

    .. [1] M. Brier, "Loss of Intranetwork and Internetwork Resting State Functional Connections with Alzheimer's
       Disease Progression" The Journal of Neuroscience, 2012.
    .. [2] P. Wang, "Aberrant intra- and inter-network connectivity architectures in Alzheimer's disease and mild
       cognitive impairment", Nature Publishing Group, 2015.

    """

    # Fetch all the network information from a excel file
    network_dict = atlas.fetch_atlas_functional_network(atlas_excel_file=atlas_file,
                                                        sheetname=sheetname,
                                                        network_column_name=network_column_name)
    # The list of network name
    network_labels_list = list(network_dict.keys())

    # Find all possible pair of network
    network_possible_pairs = list(itertools.combinations(network_labels_list, 2))
    subjects_inter_network_connectivity_matrices = dict.fromkeys(groupes)
    n_network = len(network_labels_list)

    # Build the network labels order from network possible
    # pair list for plotting purpose
    network_possible_pairs_array = np.array(network_possible_pairs)
    network_labels_order = remove_duplicate(seq=list(network_possible_pairs_array[:, 1]))
    network_labels_order.insert(0, network_possible_pairs[0][0])

    # Initialize a dictionnary for each network
    # Extract the inter network connectivity matrix
    for groupe in groupes:
        subjects_list = list(subjects_individual_matrices_dictionnary[groupe].keys())
        subjects_inter_network_connectivity_matrices[groupe] = dict.fromkeys(subjects_list)
        for subject in subjects_list:
            subjects_inter_network_connectivity_matrices[groupe][subject] = dict.fromkeys(kinds)
            for kind in kinds:
                subjects_inter_network_connectivity_matrices[groupe][subject][kind] = \
                    dict.fromkeys(network_possible_pairs)
                all_inter_network_strength = []
                for pair_of_network in network_possible_pairs:
                    inter_network_coefficient_to_sum = []
                    inter_network_coefficient_to_sum_mask = []
                    # Fetch the roi index in 4D atlas of the first network
                    # in the pair via the network dictionary, and number of rois
                    first_network_roi_index = network_dict[pair_of_network[0]]['dataframe'][roi_indices_column_name]
                    # Fetch the roi index in 4D atlas of the second network in the pair
                    second_network_roi_index = network_dict[pair_of_network[1]]['dataframe'][roi_indices_column_name]
                    # Loop over the pair of coefficient between the pair of network
                    for j in first_network_roi_index:
                        for i in second_network_roi_index:
                            # Append the coefficient to sum of all possible inter network coefficient pair
                            inter_network_coefficient_to_sum.append(
                                subjects_individual_matrices_dictionnary[groupe][subject][kind][i, j])
                            # Append the corresponding values in the masked array accounting for discarded rois.
                            inter_network_coefficient_to_sum_mask.append(
                                subjects_individual_matrices_dictionnary[groupe][subject]['masked_array'][i, j])
    
                    # Construct the masked array structure
                    # TODO: ABS or not Abs?
                    inter_network_pair_masked_array = np.ma.array(data=inter_network_coefficient_to_sum,
                                                                  mask=inter_network_coefficient_to_sum_mask)
                    # Average of the masked array structure, this is the inter network strength
                    inter_network_strength = inter_network_pair_masked_array.mean()
    
                    all_inter_network_strength.append(inter_network_strength)
    
                    subjects_inter_network_connectivity_matrices[groupe][subject][kind][pair_of_network] = {
                        'strength': inter_network_strength,
                        'connectivity coefficient': inter_network_pair_masked_array.data,
                        'masked array': inter_network_pair_masked_array.mask
                                                                                                            }
    
                subjects_inter_network_connectivity_matrices[groupe][subject][kind] = \
                    vec_to_sym_matrix(np.array(all_inter_network_strength), diagonal=np.ones(n_network)/sqrt(2))

    return subjects_inter_network_connectivity_matrices, network_labels_order


def mean_of_flatten_connectivity_matrices(subjects_individual_matrices_dictionary, groupes, kinds):
    """Return the flat mean connectivity for each subjects.

    Parameters
    ----------
    subjects_individual_matrices_dictionary: dict
        The subjects connectivity dictionnary containing connectivity matrices
        and corresponding mask array for discarded rois, for each group.
    groupes: list
        The list of groupes in the study
    kinds: list
        The list of kinds in the study

    Returns
    -------
    output: dict
        A dictionnary containing for each subject in group, a masked array structure containing
        the numerical array of the connectivity coefficient of interest, along with the boolean
        mask accounting for discarded rois. The second key, contain the mean of the extracted
        connectivity coefficient for each subject, also accounting for discarded rois it they
        exist.

    Notes
    -----
    The subjects connectivity matrices shouldn't be vectorized, the shape should
    be (n_features, n_features).
    """
    connectivity_of_interest = dict.fromkeys(list(subjects_individual_matrices_dictionary.keys()))
    for groupe in groupes:
        subjects_list = subjects_individual_matrices_dictionary[groupe].keys()
        connectivity_of_interest[groupe] = dict.fromkeys(list(subjects_list))
        for subject in subjects_list:
            connectivity_of_interest[groupe][subject] = dict.fromkeys(kinds)
            for kind in kinds:
                # Vectorize the array for the current kind
                diag_connectivity, vectorize_connectivity = array_operation.vectorizer(
                    numpy_array=subjects_individual_matrices_dictionary[groupe][subject][kind],
                    discard_diagonal=True,
                    array_type='numeric')
                # Vectorize the corresponding boolean mask
                diag_mask, vectorize_mask = array_operation.vectorizer(
                    numpy_array=subjects_individual_matrices_dictionary[groupe][subject]['masked_array'],
                    discard_diagonal=True,
                    array_type='boolean')
                # Compute the masked mean accounting for discarded ROIs
                vectorize_connectivity_masked = np.ma.masked_array(data=vectorize_connectivity,
                                                                   mask=vectorize_mask)
                flat_mean_subject_connectivity = vectorize_connectivity_masked.mean()
                # Fill the dictionnary with the mean connectivity of interest for the current subjects
                connectivity_of_interest[groupe][subject][kind] = {
                    'connectivity masked array': vectorize_connectivity_masked,
                    'mean connectivity': flat_mean_subject_connectivity}

    return connectivity_of_interest


def tangent_space_projection(reference_group, group_to_project, bootstrap_number,
                             bootstrap_size, output_directory, verif_null=True,
                             correction_method="bonferroni",
                             alpha=0.05, statistic="t"):
    """Project a group of time series to a space tangent to a reference connectivity matrix.
    The reference matrix is derived from a reference times series group. Then, each edges of
    the projected matrices is tested regarding the reference group through a one sample t-test
    or a simple z-score. The null distribution of edges is derived with bootstrap.
    Please, take a look to the reference paper for further a deeper understanding of the
    method.

    Parameters
    ----------
    reference_group: numpy.array
        The stack of the vectorized connectivity matrices of shape
        (n_subjects, n_features). This is the reference point where
        the tangent space will be computed. All tangent connectivity
        matrices will be derived regarding this reference set of matrices.
    group_to_project: numpy.array
        The stack of the vectorized connectivity matrices of shape
        (n_subjects, n_features). Each connectivity matrices will be
        projected in the tangent space, at the reference matrix of the
        reference group.
    bootstrap_number: int
        The null distribution of each edges is derived trough bootstrapping
        of the reference set of matrices. Usually around 500.
    bootstrap_size: int
        The size of the bootstrapped sample, i.e the number of
        subjects to pick from the reference set.
    output_directory: str
        The full path of the directory for saving the results
    verif_null: bool, optional
        If True, generate a report choosing a set of 20 random
        rois and plot the null distributions in those rois. True
        by defaults
    correction_method: str, optional
        The correction method used for to correct for
        multiple comparison. The default is the classical
        bonferroni method. You can choose among all the correction
        proposed by the statsmodels library.
    alpha: float, optional
        The type I error rate used for the multiple comparison
        correction. Set to .05 by defaults.
    statistic: str, optional
        The statistic used, a one sample t-test by default. Other choices
        are: z, for a simple z-score.

    Returns
    -------
    output: dict
        A dictionary with the following fields:
            - null_distribution: The array containing
            the estimated null distribution for each edges.
            - p_values_corrected: The p values, adjusted
            after the multiple comparison correction
            - reference_group_tangent_mean: The reference
            matrix, derived from the mean of the matrices
            from the reference group. This is the point
            where all matrices are projected.
            - reference_group_tangent_matrices: The connectivity
            matrices in the tangent space for the reference
            group.
            - group_to_project_tangent_matrices: The
            connectivity matrices in the tangent space
            for the projected group
            - group_to_project_stats: The chosen
            statistic for each subject, each edges
            in the projected set of subjects.

    References
    ----------

    .. [1] G. Varoquaux et al. “Detection of brain functional-connectivity
        difference in post-stroke patients using group-level covariance modeling", MICCAI 2010
    """
    # Initialize a list containing the null distribution
    null_distribution = []

    # Generate a bootstrap matrix of indices
    # before the loop.
    indices = np.arange(0, reference_group.shape[0], step=1)
    bootstrap_matrix = np.array([np.random.choice(
        indices,
        size=bootstrap_size,
        replace=False) for b in range(bootstrap_number)])

    # Loop over the bootstrap indices
    for b in range(bootstrap_number):
        print('Bootstrap # {}'.format(b))
        bootstrap_reference_group = bootstrap_matrix[b, :]
        # We chose one subject of the reference group
        # among the bootstrap sample
        left_out_subject = bootstrap_reference_group[0]
        subset_reference_group = bootstrap_reference_group[1:]
        # Compute mean matrices, and connectivity matrices
        # in the tangent on the subset of subject from the
        # reference group without the leftout subject.
        connectivity_measure = ConnectivityMeasure(kind='tangent',
                                                   vectorize=True,
                                                   discard_diagonal=True)
        reference_group_subset_matrices = \
            connectivity_measure.fit_transform(X=reference_group[subset_reference_group, ...])
        reference_group_subset_mean = connectivity_measure.mean_
        # Project at the previously computed mean
        # the leftout control subject
        leftout_subject = \
            connectivity_measure.transform(
                X=[reference_group[left_out_subject, ...]])[0]

        # Compute a z-score between the left out controls tangent connectivity
        # and the resampled controls group
        if statistic == "z":

            stat_null = zmap(scores=leftout_subject, compare=reference_group_subset_matrices)
        elif statistic == "t":

            # Another statistic: one sample t test, controls - the left out controls
            # as contrast
            stat_null = -1.0 * ttest_1samp(a=reference_group_subset_matrices, popmean=leftout_subject)[0]
        else:
            raise ValueError("Statistic type unknown. Type are z or t statistic, "
                             "you entered {}".format(statistic))

        # Store the results of the test, as null distribution
        null_distribution.append(stat_null)

    null_distribution_array = np.array(null_distribution)
    if verif_null:
        # plot of some null distribution randomly chosen
        n_rois = 20
        random_roi = np.random.choice(a=np.arange(start=0, stop=null_distribution_array.shape[1], step=1),
                                      size=n_rois)
        with PdfPages(os.path.join(output_directory, "null_distribution_random_roi.pdf")) \
                as pdf:

            for i in range(n_rois):
                plt.subplot(10, 2, i + 1)
                plt.hist(x=null_distribution_array[..., random_roi[i]], bins='auto', edgecolor='black')
                plt.title('# {}'.format(random_roi[i]))
                plt.subplots_adjust(hspace=1, wspace=.01)
            pdf.savefig()
            plt.show()

    # Compute tangent connectivity matrices on the whole
    # reference group
    reference_group_connectivity = ConnectivityMeasure(kind='tangent',
                                                       vectorize=True,
                                                       discard_diagonal=True)

    reference_group_tangent_matrices = reference_group_connectivity.fit_transform(X=reference_group)
    # Compute the tangent mean for the controls group
    reference_group_tangent_mean = reference_group_connectivity.mean_

    # Project each subject's matrices belonging to
    # group to project in the tangent space
    group_to_project_tangent_matrices = reference_group_connectivity.transform(
        X=group_to_project)
    save_object(
        object_to_save=group_to_project_tangent_matrices,
        saving_directory=output_directory,
        filename='projected_group_connectivity_matrices')

    # Test for each projected subject, and for each brain connection
    # if there exist a significant connectivity difference between the
    # a subject and the reference group.
    if statistic == "z":
        # Z-statistic between a projected subject and the whole
        # reference group
        group_to_project_scores = zmap(scores=group_to_project_tangent_matrices,
                                       compare=reference_group_tangent_matrices)
    elif statistic == "t":
        # A one sample t-test between the reference group and a projected subject
        group_to_project_scores = np.array([-1.0 * ttest_1samp(
            a=reference_group_tangent_matrices, popmean=group_to_project_tangent_matrices[p, ...])[0]
                                            for p in range(group_to_project_tangent_matrices.shape[0])])
    else:
        raise ValueError("Statistic type unknown. Type are z or t statistic, "
                         "you entered {}".format(statistic))

    # Compute the raw p value based on the previously
    # computed null distribution, for each subject.
    p_values = np.empty(shape=group_to_project_scores.shape)
    for subject in range(group_to_project_scores.shape[0]):
        for i in range(group_to_project_scores.shape[1]):
            # with the z-score, or the t-score
            p_values[subject, i] = \
                (np.sum(np.abs(null_distribution_array[:, i]) > np.abs(group_to_project_scores[subject, i])) + 1) / \
                (bootstrap_number + 1)

    # correct the p values for each subject
    # in the group to project
    p_values_corrected = np.empty(shape=group_to_project_scores.shape)
    for subject in range(group_to_project_scores.shape[0]):
        p_values_corrected[subject, ...] = multipletests(pvals=p_values[subject, ...],
                                                         method=correction_method,
                                                         alpha=alpha)[1]

    return {"null_distribution": null_distribution_array, "p_values_corrected": p_values_corrected,
            "reference_group_tangent_mean": reference_group_tangent_mean,
            "reference_group_tangent_matrices": reference_group_tangent_matrices,
            "group_to_project_tangent_matrices": group_to_project_tangent_matrices,
            "group_to_project_stats": group_to_project_scores}



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 18 10:10:31 2013

@author: jinpeng.li@cea.fr
"""

import numpy as np
from epac.map_reduce.reducers import Reducer
from epac.map_reduce.results import Result
from epac.workflow.base import key_pop


class PValR2Reducer(Reducer):
    """Reducer that computes p-values of stattistics.

    """
    def __init__(self):
        self.pattern = "Perm\(nb=([a-zA-Z0-9]+)(.*)"

    def get_diff_perm_nbs(self, result):
        import re
        import numpy as np
        res_keys = result.keys()
        ret_diff_perm_nbs = set()
        for res_key in res_keys:
            re_res = re.search(self.pattern, res_key)
            if re_res:
                ret_diff_perm_nbs.add(int(re_res.group(1)))
        return list(ret_diff_perm_nbs)

    def get_max_r2_with_perm_nb(self, result, perm_nb):
        import re
        import numpy as np
        res_keys = result.keys()
        ret_max_r2 = None
        for res_key in res_keys:
            re_res = re.search(self.pattern, res_key)
            if re_res:
                perm_nb_i = re_res.group(1)
                if perm_nb_i == repr(perm_nb):
                    cur_max = np.max(result[res_key]['r2'])
                    if not ret_max_r2:
                        ret_max_r2 = cur_max
                    elif ret_max_r2 < cur_max:
                        ret_max_r2 = cur_max
        return ret_max_r2

    def reduce(self, result):
        diff_perm_nbs = self.get_diff_perm_nbs(result)
        max_r2 = {}
        for perm_nb in diff_perm_nbs:
            max_r2[perm_nb] = self.get_max_r2_with_perm_nb(result,
                                                           perm_nb)
        r2_no_perms = max_r2[0]
        count = 0
        for i in max_r2:
            if i == 0:
                continue
            if r2_no_perms < max_r2[i]:
                count += 1
        p_value = float(count) / float(len(max_r2))
        _, res_key = key_pop(result.keys()[0], index=-1)
        out = Result(key=res_key)
        out["pval"] = p_value
        return out



import numpy as np
from patsy import dmatrix
from sklearn.model_selection import LeaveOneOut
from scipy import stats
import pandas as pd
from conpagnon.machine_learning.CPM_method import predictors_selection_linear_model, fit_model_on_training_set, \
    compute_summary_subjects_summary_values, predictors_selection_correlation, predictor_selection_pcorrelation


def predict_behavior(vectorized_connectivity_matrices, behavioral_scores,
                     selection_predictor_method='correlation',
                     significance_selection_threshold=0.01,
                     confounding_variables=None, confounding_variables_kwarg=None):
    """The Connectome Predictive Modelling pipeline. This function select the predictors,
    train/test a linear model on the selected predictors following a Leave One Out cross
    validation scheme.

    Parameters
    ----------
    vectorized_connectivity_matrices: numpy.array of shape (n_subjects, n_features)
        The stack of the vectorized (lower or upper triangle of the connectivity matrices)
        connectivity matrices. Be careful, the matrices should be stack in the same order
        as the vector of scores to predict !
    behavioral_scores: numpy.array of shape (n_subject, 1)
        The vector of scores to predict. The scores should be in the same
        order as the vectorized connectivity matrices stack.
    selection_predictor_method: str, optional
        The predictors selection method. By default, a correlation between
        each connectivity coefficient and scores is computed, and the resulted
        correlation matrices is threshold at a type I error rate equal to 0.01.
        Other selection are available: 'linear_model', 'partial correlation'.
    significance_selection_threshold: float, optional
        The significance threshold during the selection procedure. By default,
        set to 0.01.
    confounding_variables: list, optional
        A list of the possible confounding variables you might
        want to add, during the selection procedure only.
    confounding_variables_kwarg: dict, optional
        A dictionary with a field called 'file_path'. This field
        should contains the full path to a file containing as
        many columns as confounding variable.

    Returns
    -------
    output 1: float
        The correlation coefficient between the predicted and true scores
        from the positively correlated set of features.
    output 2: float
        he correlation coefficient between the predicted and true scores
        from the negatively correlated set of features.

    """
    # Initialize leave one out object
    leave_one_out_generator = LeaveOneOut()

    # Initialization of behavior prediction vector
    behavior_prediction_positive_edges = np.zeros(len(vectorized_connectivity_matrices.shape[0]))
    behavior_prediction_negative_edges = np.zeros(len(vectorized_connectivity_matrices.shape[0]))

    # Date preprocessing
    if confounding_variables is not None:
        # read confounding variable file
        if confounding_variables_kwarg['file_path'].endswith('csv', 'txt'):
            # Read the text file, and fetch the columns corresponding to the confounding variables
            confounding_variables_data = pd.read_csv(confounding_variables_kwarg['file_path'])[confounding_variables]
            # Construct the design matrix containing the confound variables
            confounding_variables_matrix = dmatrix(formula_like='+'.join(confounding_variables),
                                                   data=confounding_variables_data,
                                                   return_type='dataframe').drop(['Intercept'], axis=1)

        elif confounding_variables_kwarg['file_path'].endswith('xlsx'):
            confounding_variables_data = pd.read_excel(confounding_variables_kwarg['file_path'])[confounding_variables]
            # Construct the design matrix containing the confound variables
            confounding_variables_matrix = dmatrix(formula_like='+'.join(confounding_variables),
                                                   data=confounding_variables_data,
                                                   return_type='dataframe').drop(['Intercept'], axis=1)

        else:
            raise ValueError('Datafile extension unrecognized')

    for train_index, test_index in leave_one_out_generator.split(vectorized_connectivity_matrices):
        # For each iteration, split the patients matrices array in train and
        # test set using leave one out cross validation
        patients_train_set, leave_one_out_patients = \
            vectorized_connectivity_matrices[train_index], vectorized_connectivity_matrices[test_index]

        # Training set behavioral scores
        training_set_behavioral_score_ = np.zeros((patients_train_set.shape[0], 1))
        training_set_behavioral_score_[:, 0] = behavioral_scores[train_index]

        # The confounding variables, stored in an array for the training set
        training_confound_variable_matrix = confounding_variables_matrix.iloc[train_index]

        if selection_predictor_method == 'linear_model':

            # Correlation of each edge to the behavioral score for training set
            R_mat, P_mat = predictors_selection_linear_model(
                training_connectivity_matrices=patients_train_set,
                training_confound_variable_matrix=training_confound_variable_matrix,
                training_set_behavioral_score=training_set_behavioral_score_)

            # Compute summary values for both positive and negative edges model
            negative_edges_mask, positive_edges_mask, negative_edges_summary_values, positive_edges_summary_values =\
                compute_summary_subjects_summary_values(
                    training_connectivity_matrices=patients_train_set,
                    significance_selection_threshold=significance_selection_threshold,
                    R_mat=R_mat, P_mat=P_mat)

        elif selection_predictor_method == 'correlation':

            R_mat, P_mat = \
                predictors_selection_correlation(training_connectivity_matrices=patients_train_set,
                                                 training_set_behavioral_scores=training_set_behavioral_score_)

            negative_edges_mask, positive_edges_mask, negative_edges_summary_values, positive_edges_summary_values =\
                compute_summary_subjects_summary_values(
                    training_connectivity_matrices=patients_train_set,
                    significance_selection_threshold=significance_selection_threshold,
                    R_mat=R_mat, P_mat=P_mat)

        elif selection_predictor_method == 'partial correlation':

            R_mat, P_mat = predictor_selection_pcorrelation(
                training_connectivity_matrices=patients_train_set,
                training_set_behavioral_scores=training_set_behavioral_score_,
                training_set_confounding_variables=training_confound_variable_matrix)

            negative_edges_mask, positive_edges_mask, negative_edges_summary_values, positive_edges_summary_values =\
                compute_summary_subjects_summary_values(
                    training_connectivity_matrices=patients_train_set,
                    significance_selection_threshold=significance_selection_threshold,
                    R_mat=R_mat, P_mat=P_mat)
        else:
            raise ValueError('Selection method not understood')

        # Fit a linear model on the training set
        positive_edge_model_fit, negative_edge_model_fit = fit_model_on_training_set(
            negative_edges_summary_values=negative_edges_summary_values,
            positive_edges_summary_values=positive_edges_summary_values,
            training_set_behavioral_score=training_set_behavioral_score_)

        # Test the positive edges model on the left out subject
        test_subject_positive_edges_summary = np.sum(np.multiply(leave_one_out_patients[0, :], positive_edges_mask))
        test_subject_negative_edges_summary = np.sum(np.multiply(leave_one_out_patients[0, :], negative_edges_mask))

        # Fit the model of on the left out subject
        behavior_prediction_negative_edges[test_index] = \
            negative_edge_model_fit.params[1]*test_subject_negative_edges_summary + \
            negative_edge_model_fit.params[0]

        behavior_prediction_positive_edges[test_index] = \
            positive_edge_model_fit.params[1]*test_subject_positive_edges_summary + positive_edge_model_fit.params[0]

    # Compare prediction and true behavioral score
    R_predict_negative_model, _ = \
        stats.pearsonr(x=behavior_prediction_negative_edges,
                       y=np.array(behavioral_scores))

    R_predict_positive_model, _ = \
        stats.pearsonr(x=np.array(behavioral_scores),
                       y=behavior_prediction_positive_edges)

    return R_predict_positive_model, R_predict_negative_model





import importlib
from conpagnon.utils import pre_preprocessing, folders_and_files_management
from conpagnon.utils import array_operation
from conpagnon.utils.array_operation import vectorizer
from scipy.stats import mstats, norm, ttest_ind, pearsonr, ttest_rel
from nilearn.connectome import sym_matrix_to_vec, vec_to_sym_matrix
import numpy as np
from statsmodels.sandbox.stats.multicomp import multipletests
import statsmodels.api as sm
import warnings
from conpagnon.computing import compute_connectivity_matrices as ccm
from patsy import dmatrix, dmatrices
import pandas as pd
from conpagnon.pylearn_mulm import mulm
import os
import errno
import copy
from scipy import linalg
import statsmodels.formula.api as smf
importlib.reload(pre_preprocessing)
importlib.reload(array_operation)
importlib.reload(ccm)
importlib.reload(mulm)

# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct  9 16:17:21 2017
@author: db242421 (dhaif.bekha@cea.fr)
Parametric test used in a resting state group connectivity analysis.
"""


def two_samples_t_test(subjects_connectivity_matrices_dictionnary, groupes, kinds, contrast,
                       preprocessing_method='fisher',
                       alpha=.05, multicomp_method='fdr_bh'):
    """Perform two samples t-test on connectivity matrices to detect group differences in connectivity
    using different kinds.

    The t-test account for discarded rois you might want to exclude in the analysis.

    Parameters
    ----------
    subjects_connectivity_matrices_dictionnary : dict
        A multi-levels dictionnary organised as follow :
            - The first keys levels is the different groupes in the study.
            - The second keys levels is the subjects IDs
            - The third levels is the different kind matrices
            for each subjects, a 'discarded_rois' key for the
            discarded rois array index, a 'masked_array' key containing
            the array of Boolean of True for the discarded_rois index, and False
            elsewhere.

    groupes : list
        The list of the two groups to detect group differences.

    kinds : list
        The list of metrics you want to perform the group comparison.
        Choices are : 'correlation', 'covariances', 'tangent', 'partial correlation',
        'precision'.
    preprocessing_method : string, optional
        The type of preprocessing methods to apply of connectivity coefficients
        of type 'correlation', 'partial correlation', 'covariances', 'precision'.
        Choices are : 'fisher'.
    contrast : list, optional
        The contrast you want to compute in the t-test. Default is [1.0, -1.0]
        to compute mean(groupes[0]) - mean(groupes[1]).
        The other contrast is [-1.0, 1.0] for mean(groupes[1]) - mean(groupes[0]).
    alpha : float, optional
        The false positive proportion, commonly named the alpha level.
        Default is 0.05.
    multicomp_method : str, optional
        The inference method for accounting the multiple comparison problems.
        Default is the classic False Discovery Rate (FDR) proposed by
        Benjamini & Hochberg, see Notes.


    Returns
    -------
    output : dict
        A dictionnary containing multiple keys :
            - 'tstatistic' : The raw statistic t-map for the chosen contrast
            2D numpy array of shape (number of regions, number of regions).

            - 'uncorrected pvalues' : The raw pvalues, 2D numpy array of shape
            (number of regions, number of regions).

            - 'corrected pvalues' : The corrected pvalues with the chosen methods.
            2D numpy array of shape (number of regions, number of regions).

            - 'significant edges' : The significant t-values after masking
            for the non significant pvalues at alpha level. 2D numpy array of shape
            (number of regions, number of regions).

            - 'significant pvalues' : The significant pvalues at level alpha.
            2D numpy array of shape (number of regions, number of regions)

            - 'significant mean effect' : The differences of mean connectivity
            between the two groups according to the chosen contrast, and non-significant
            connexion are mask at alpha level. 2D numpy array of shape
            (number of regions, number of regions).

    Raises
    ------
    ValueError : If the number in groupes is strictly less than 2 raise
    a ValueError, print a warning and take the two first groupes in the list
    else.

    ValueError : If the contrast is an unrecognized contrast is entered.

    See Also
    --------

    compute_connectivity_matrices.individual_connectivity_matrices :
        This is the function which compute the connectivity matrices for
        the chosen kinds, and return a structured dictionnary with can be used
        for the argument `subjects_connectivity_matrices_dictionnary`.

    pre_preprocessing.fisher_transform :
        Compute the Fisher Transform of connectivity coefficient

    Notes
    -----
    The two sample t-test I perform here account for masked array using
    statistical test for masked array in the Scipy packages. That is,
    if in the subjects masked_array, some regions are masked, they will be
    discarded when performing the t-test. If the mask contain only False
    value, no region are discarded.

    All the metric are symmetric, therefore I only compute the t-test
    for the lower part of the matrices of shape n_columns * (n_columns + 1) /2

    The correction for the multiple comparison is applied to the resulting
    t-values, and entire matrices are reconstructed after and resulting shape
    is (number of regions, number of regions)

    Different method for accounting the multiple comparison problems exist,
    please refers to the corresponding function in the statsmodels library :
    statsmodels.sandbox.stats.multicomp.multipletests.

    For the tangent kinds, keep in mind that the mean effect is computed
    in the tangent space only, that is, not in the same space as classical
    metrics like covariances, correlation of partial correlation.

    """

    stacked_matrices = pre_preprocessing.stacked_connectivity_matrices(subjects_connectivity_matrices=
                                                                       subjects_connectivity_matrices_dictionnary,
                                                                       kinds=kinds)

    # Initialise a dictionnary for saving the t_statistic, and the uncorrected and corrected p-value
    t_test_dictionnary = dict.fromkeys(kinds)

    # Check if they are at least two samples:
    n_groupes = len(groupes)

    if n_groupes < 2:
        raise ValueError('Two samples t-test requires at '
                         'least two samples...only found {} group'.format(n_groupes))
    elif n_groupes > 2:
        warnings.warn('{} groups was found ! Only the first two groups will be use '
                      'for the two sample t-test: {} and {}'.format(n_groupes,
                                                                    groupes[0],
                                                                    groupes[1]))

    if contrast == [1.0, -1.0]:
        print('Computing two sample t-test for kinds {} and contrast {} - {}'.format(kinds, groupes[0], groupes[1]))
    elif contrast == [-1.0, 1.0]:

        print('Computing two sample t-test for kinds {} and contrast {} - {}'.format(kinds, groupes[1], groupes[0]))
    else:
        raise ValueError('Unrecognized contrast, only [1.0,1.0] or [-1.0,-1.0] are accepted. '
                         'You enter {}'.format(contrast))

    if 'tangent' in kinds:
        warnings.warn('I think using a two sample t-test in this fashion on tangent '
                      'space should be interpreted carefully !')

    # Fisher transform coefficients
    for kind in kinds:
        if preprocessing_method == 'fisher':
            if kind != 'tangent':
                X = pre_preprocessing.fisher_transform(symmetric_array=stacked_matrices[groupes[0]][kind])
                Y = pre_preprocessing.fisher_transform(symmetric_array=stacked_matrices[groupes[1]][kind])
            elif kind == 'tangent':
                X = stacked_matrices[groupes[0]][kind]
                Y = stacked_matrices[groupes[1]][kind]
        elif preprocessing_method == 'already_preprocessed':
            X = stacked_matrices[groupes[0]][kind]
            Y = stacked_matrices[groupes[1]][kind]
        else:
            raise ValueError('Unrecognized pre-processing method')

        # Vectorize the connectivity matrices, discarding the diagonal
        X_vectorize = sym_matrix_to_vec(symmetric=X)
        Y_vectorize = sym_matrix_to_vec(symmetric=Y)
        # Vectorize the corresponding boolean array
        vec_X_mask = array_operation.vectorize_boolean_mask(
            symmetric_boolean_mask=stacked_matrices[groupes[0]]['masked_array'])
        vec_Y_mask = array_operation.vectorize_boolean_mask(
            symmetric_boolean_mask=stacked_matrices[groupes[1]]['masked_array'])

        # Create a numpy masked array structure for the two sample
        X_ = np.ma.array(data=X_vectorize, mask=vec_X_mask)
        Y_ = np.ma.array(data=Y_vectorize, mask=vec_Y_mask)

        # Finally perform a two sample t-test for masked array along the first
        # dimension according contrast, accounting for discarded rois
        if contrast == [1.0, -1.0]:
            t_stats_vec, pvalues_vec = mstats.ttest_ind(a=X_, b=Y_, axis=0)
        elif contrast == [-1.0, 1.0]:
            t_stats_vec, pvalues_vec = mstats.ttest_ind(a=Y_, b=X_, axis=0)

        # Replace nan values in pvalues_vec by 1 to avoid failure in correction method
        pvalues_vec[np.isnan(pvalues_vec)] = 1.

        # Correction of pvalues, reject of null hypotheses below alpha level.
        reject_, pvalues_vec_corrected, _, _ = multipletests(pvals=pvalues_vec.data,
                                                             alpha=alpha,
                                                             method=multicomp_method)

        # Computing the effect: difference between mean according to the contrast vector:
        groupes_mean_matrices = ccm.group_mean_connectivity(
            subjects_connectivity_matrices=subjects_connectivity_matrices_dictionnary,
            kinds=kinds)

        mean_X = groupes_mean_matrices[groupes[0]][kind]
        mean_Y = groupes_mean_matrices[groupes[1]][kind]

        if contrast == [1.0, -1.0]:
            mean_effect = mean_X - mean_Y
        elif contrast == [-1.0, 1.0]:

            mean_effect = mean_Y - mean_X

        # Reconstruction of statistic for each kind
        t_stats_matrix = vec_to_sym_matrix(vec=t_stats_vec)
        pvalues_uncorrected_matrix = vec_to_sym_matrix(vec=pvalues_vec)
        pvalues_corrected_matrix = vec_to_sym_matrix(vec=pvalues_vec_corrected)

        # Fill diagonal with p = 1 for corrected pvalues
        np.fill_diagonal(pvalues_corrected_matrix, 1)

        # Reconstruction of significant brain connection and significant pvalues
        reject_boolean_matrix = vec_to_sym_matrix(vec=reject_)
        # Fill diagonal with 0 in the reconstructed numerical rejection array
        np.fill_diagonal(reject_boolean_matrix, 0)

        significant_edges = np.multiply(t_stats_matrix, reject_boolean_matrix)
        significant_pvalues = np.multiply(pvalues_corrected_matrix, reject_boolean_matrix)
        significant_mean_effect = np.multiply(mean_effect, reject_boolean_matrix)
        uncorrected_mean_effect = np.multiply(mean_effect, pvalues_uncorrected_matrix < alpha)

        # Save the t_stats_matrix, the uncorrected p values matrix, the corrected pvalues matrix in the dictionary.
        t_test_dictionnary[kind] = {'tstatistic': t_stats_matrix,
                                    'uncorrected pvalues': pvalues_uncorrected_matrix,
                                    'corrected pvalues': pvalues_corrected_matrix,
                                    'significant edges': significant_edges,
                                    'significant pvalues': significant_pvalues,
                                    'significant mean effect': significant_mean_effect,
                                    'total mean effect': mean_effect,
                                    'uncorrected mean effect': uncorrected_mean_effect,
                                    'tested_contrast': contrast}

    return t_test_dictionnary


def linear_regression(connectivity_data, data, formula, NA_action,
                      kind, subjects_to_drop=None, sheetname=0, save_regression_directory=None,
                      contrasts='Id', compute_pvalues=True, pvalues_tail='two_tailed',
                      alpha=0.05, pvals_correction_method=['fdr_bh'], nperms_maxT=10000,
                      vectorize=False, discard_diagonal=False):
    # TODO: add a way to select column containing the subjects ID, and
    # TODO: shift it to be the index of the DataFrame
    """Fit a linear model on connectivity coefficients across subjects.

    Parameters
    ----------
    connectivity_data: dict
        The connectivity matrix organised in a dictionary. The subject identifier
        as keys, and metric as values, i.e the matrix which contain the connectivity coefficient.
        As metric are symmetric, matrix as to be vectorize in a 1D array of shape n_features.
    formula: string
        The model, in a R fashion.
    data: string
        The full path to the xlsx data file, containing all the dependant variables
        in the model you want to estimate.
    NA_action: string
        Directive for handling missing data in the xlsx file. Choices are :
        'drop': subject will discarded in the analysis, 'raise': raise an
        error if missing data is present.
    sheetname: int
        The position in your excel file of the sheet of interest.
    subjects_to_drop: list, optional
        List of subjects you want to discard in the analysis. If None, all the
        row in the dataframe are kept. Default is None.
    kind: string
        The metric, present in the provided connectivity data you want to perform analysis.
    save_regression_directory: string
        The full path to a directory for saving the regression results
    contrasts: string or numpy.array of shape (n_features,n_features) optional
        The contrast vector for infering the regression coefficients.
        Default is 'Id', all regressors are tested.
    compute_pvalues: bool, optional
        If True pvalues are computed. Default is True
    pvalues_tail: string, optional
        If 'two_tailed', a two-sided t-test is computed. If 'one_tailed' a one
        tailed t-test is computed.
    alpha: float, optional
        The Type error rate. Corrected p-values above alpha will be discarded.
        Default is 0.05.
    pvals_correction_method: string, optional
        The method for accounting for the multiple comparison problems.
        Choices are among the statsmodels library : {'bonferroni', 'sidak', 'holm-sidak', 'holm',
        'simes-hochberg', 'hommel', 'fdr_bh', 'fdr_by', 'fdr_tsbh', 'fdr_tsbky'}, and the
        'maxT' method in the mulm library. Default is 'fdr_bh'.
    nperms_maxT: int, optional
        If maximum statistic correction is chosen, the number of permutations. Default is 10000.

    Returns
    -------
    output 1: dict
        The regression results, with the regressors variable as keys, and
        corrected and corrected p-values matrix, significant t-values matrix,
        t-values matrix under the alpha uncorrected threshold.
    output 2: numpy.array of shape (n_samples, n_features)
        The design matrix of the analysis.
    output 3: numpy.array of shape (n_samples, q), q: number of model to fit.
        The independent variable matrix with all the multiple outcome to fit.

    """
    # Reading the data
    df = pd.read_excel(data, sheet_name=sheetname)
    # Set the index to the subjects identifier column
    df = df.set_index(['subjects'])
    
    # Drop the subjects we want to discard :
    df_c = df.copy()
    if subjects_to_drop is not None:
        df = df_c.drop(labels=subjects_to_drop)
    else:
        df = df_c
    
    # Build the design matrix according the dataframe and regression model.
    X_df = dmatrix(formula_like=formula, data=df, return_type='dataframe',
                   NA_action=NA_action)

    # Stacked vectorized connectivity matrices in the same order of subjects
    # index list of the DESIGN MATRIX, because of missing data, not all subjects
    # will be in the analysis.
    # All the subjects present in the excel file
    general_regression_subjects_list = X_df.index
    # Intersection of subjects to perform regression and the general list because
    # of possible dropped NA values.
    regression_subjects_list = \
        list(set(connectivity_data.keys()).intersection(general_regression_subjects_list))
    y = np.array([connectivity_data[subject][kind] for subject in regression_subjects_list])
    
    if vectorize:
        y = sym_matrix_to_vec(y, discard_diagonal=discard_diagonal)
    else:
        y = y
    
    # Conversion of X_df into a classic numpy array
    X_df = X_df.loc[regression_subjects_list]
    X = np.array(X_df.loc[regression_subjects_list])
    
    # Setting the contrast vector
    if contrasts == 'Id':
        contrasts = np.identity(X.shape[1])
    else:
        contrasts = contrasts
    
    # Mass univariate testing using MUOLS library
    mod = mulm.MUOLS(Y=y, X=X).fit()
    raw_tvals, raw_pvals, dfree = mod.t_test(contrasts=contrasts,
                                             pval=compute_pvalues,
                                             two_tailed=pvalues_tail)
    
    # Compute prediction of the models
    y_prediction = mod.predict(X=X)
    
    # Replace nan values in pvalues_vec by 1 to avoid failure in correction method
    raw_pvals[np.isnan(raw_pvals)] = 1.
    
    # Initialize boolean mask for rejected H0 hypothesis, i.e for corrected pvalues < alpha.
    reject_ = np.zeros(raw_pvals.shape, dtype='bool')
    # Initialize array to save corrected pvalues
    pvalues_vec_corrected = np.zeros(raw_pvals.shape)
    
    # Save the results dictionary per correction method
    correction_method_regression_results = dict.fromkeys(pvals_correction_method)
    
    # Saving the results of the regression
    # Creation of the directory containing the regression results
    if save_regression_directory is not None:
        try:
            os.makedirs(save_regression_directory)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        
    # Choose multiple comparison correction method among the statsmodels library
    for corr_method in pvals_correction_method:
        
        # Listing the dependent variable, in the order of the output of
        # the regression analysis
        covariable_name = X_df.columns
        # Save a dictionnary with the principal
        regression_results = dict.fromkeys(covariable_name)
        
        if corr_method in ['fdr_bh', 'bonferroni']:
            
            # Flatten the array of raw p-values
            raw_p_shape = raw_pvals.shape
            # Call fdr correction, and reshape the results to (n_variable, n_features)
            pvalues_vec_corrected = multipletests(pvals=raw_pvals.flatten(),
                                                   method=corr_method, alpha=alpha)[1].reshape(raw_p_shape)
            
            reject_ = multipletests(pvals=raw_pvals.flatten(),
                                                   method=corr_method, alpha=alpha)[0].reshape(raw_p_shape)
        
        elif corr_method == 'maxT':
            # Inference with the maximum statistic method :
            _, pvalues_vec_corrected, _, null_distribution = \
            mod.t_test_maxT(contrasts=contrasts,
                            nperms=nperms_maxT)
            # Manual construction for the reject_ mask:
            reject_ = pvalues_vec_corrected < alpha
        
        else:
            raise ValueError('Unrecognized correction method ! \n'
                             'Please refer to the docstring function.')
        
        # Reconstruction in a matrix structure of the different output
        
        # Reconstruction of boolean reject_ mask : it's a binary mask
        reject_m_ = vec_to_sym_matrix(reject_)
        # Reconstruction of raw t values
        raw_tvals_m = vec_to_sym_matrix(raw_tvals)
        # Reconstruction of raw p values
        raw_pvals_m = vec_to_sym_matrix(raw_pvals)
        # Reconstruction of corrected p values
        corrected_pvals_m = vec_to_sym_matrix(pvalues_vec_corrected)
        # Construction of masked raw t values according to corrected p values
        # under alpha threshold
        significant_tvals_m = np.multiply(raw_tvals_m, reject_m_)

        for i in range(len(X_df.columns)):
            regression_results[covariable_name[i]] = \
                {'raw pvalues': raw_pvals_m[i, :, :],
                 'raw tvalues': raw_tvals_m[i, :, :],
                 'corrected pvalues': corrected_pvals_m[i, :, :],
                 'significant tvalues': significant_tvals_m[i, :, :]}

        correction_method_regression_results[corr_method] = {'results': regression_results}
    
        if corr_method == 'maxT':    
            correction_method_regression_results['maximum T null distribution'] = null_distribution

    # saving the dictionary
    if save_regression_directory is not None:
        folders_and_files_management.save_object(correction_method_regression_results,
                                                 save_regression_directory,
                                                 'regression_results.pkl')

    return correction_method_regression_results, X_df, y, y_prediction, regression_subjects_list


def functional_connectivity_distribution_estimation(functional_connectivity_estimate):
    # TODO : When functional_connectivity_estimate contain nan, the estimation fail,
    # TODO: I have to add a case for data contain non finite value
    """Estimates the mean and standard deviation of functional connectivity distribution assuming a Gaussian behavior.

    Parameters
    ----------
    functional_connectivity_estimate: numpy.array, shape(n_features, n_features)
    or shape 0.5*n_features*(n_features + 1)
        A functional connectivity matrices, if a 2D array is provided it will vectorized discarding the diagonal.

    Returns
    -------
    output 1: numpy.array shape 0.5*n_features*(n_features + 1)
        The vectorized functional connectivity array.
    output 2: float
        The estimated mean of the data.
    output 2: float
        The estimated standard deviation of the data.

    See Also
    --------
    scipy.stats.norm :
        This function from the scipy library is used here,
        to estimate the mean and the standard deviation of the data.
    pre_preprocessing.fisher_transform :
        To ensure a normal behavior of the connectivity coefficient,
        this function apply a classical Fisher transform to the data.

    Notes
    -----
    We assume here that the functional connectivity your're dealing with have a **Gaussian** behavior,
    and therefore can be describe properly with two parameters: the mean and the standard deviation.
    To ensure a Gaussian behavior transformation of the connectivity coefficient should be used like
    Fisher transform for correlation or partial correlation for example.

    """

    # Check the dimension of the passed functional matrices
    functional_connectivity_shape = functional_connectivity_estimate.shape
    # If it is a 2D matrix we vectorize it, discarding the diagonal to have a 1D distribution array
    if len(functional_connectivity_shape) == 2:
        vectorized_functional_connectivity_estimate = sym_matrix_to_vec(functional_connectivity_estimate,
                                                                        discard_diagonal=True)
    else:
        vectorized_functional_connectivity_estimate = functional_connectivity_estimate

    # We estimate the parameters of the distribution assuming it is Gaussian
    mean_data, std_data = norm.fit(vectorized_functional_connectivity_estimate)

    return vectorized_functional_connectivity_estimate, mean_data, std_data


def mean_functional_connectivity_distribution_estimation(mean_groups_connectivity_matrices):
    """Estimates for the mean connectivity matrices for each group, the mean and standard deviation
    assuming gaussian distribution

    Parameters
    ----------
    A multi-levels dictionnary organised as follow :
        - The first keys levels is the different groups in the study.
        - The second keys levels is the mean connectivity matrices for
        the different kinds. They are array of shape (number of regions , number of regions).

    Returns
    -------
    output : dict
        A dictionnary organised as follow:
            - The first keys levels is the different groups in the study.
            - The second keys levels is the kinds present in the provided dictionnary.
            - The third levels keys contain the estimated mean, the estimated standard deviation,
            and the vectorized array of connectivity coefficients.

    Notes
    -----
    Apply a Z-fisher transformation to the input matrices can be useful to improve and ensure a normal
    behavior of the data.

    """

    parameters_functional_connectivity_estimation = dict.fromkeys(mean_groups_connectivity_matrices)

    for groupe in mean_groups_connectivity_matrices.keys():
        parameters_functional_connectivity_estimation[groupe] = dict.fromkeys(mean_groups_connectivity_matrices[groupe])
        for kind in mean_groups_connectivity_matrices[groupe].keys():
            mean_kind_matrices = mean_groups_connectivity_matrices[groupe][kind]
            # Estimating the mean and standard deviation for the distribution
            vectorized_kind_matrices, kind_mean_data, kind_std_data = functional_connectivity_distribution_estimation(mean_kind_matrices)
            # Fill the dictionnary saving the mean and standard deviation for each kind, and the vectorized connectivity matrices
            parameters_functional_connectivity_estimation[groupe][kind] = {'mean': kind_mean_data, 'std': kind_std_data,
                                                                           'vectorized connectivity': vectorized_kind_matrices}

    return parameters_functional_connectivity_estimation


def two_sample_t_test_on_mean_connectivity(mean_matrix_for_each_subjects, kinds, groupes, contrast):
    """Performs a two sample t-test between the two groupes in the study on the mean matrix of each subjects.

    Parameters
    ----------
    mean_matrix_for_each_subjects: dict
        A dictionnary containing, at the first level, the groups as keys, and the kinds as values.
        Inside each each kind, a dictionnary containing the subjects identifier as values, and
        the mean functional connectivity of that subject as values.
    kinds: list
        The list of kinds to test.
    groupes: list
        The list of the two groupes in the study.
    contrast: list of int.
        A list of int, to precise the contrast between the two group. If contrast
        is set to the vector [1.0, -1.0], the computed t values is base on the contrast
        groupes[0] - groupes[1], and if contrast is [-1.0, 1.0], it's the contrary.

    Returns
    -------
    output: dict
        A dictionnary containing each kinds as keys, and the t-statistic and p-values as
        values for each kinds.

    Notes
    -----
    The p-values is based on a two-sided statistic.


    """

    t_test_dictionnary_functional_score = dict.fromkeys(kinds)
    for kind in kinds:

        # Stacked the functional score of the first group, the first in the list of groupes
        x = np.array([mean_matrix_for_each_subjects[groupes[0]][kind][s] for s in mean_matrix_for_each_subjects[groupes[0]][kind].keys()])
        # Stack the functional score of the second group, the second in the list of groupes
        y = np.array([mean_matrix_for_each_subjects[groupes[1]][kind][s] for s in mean_matrix_for_each_subjects[groupes[1]][kind].keys()])

        # Depending on the contrast
        if contrast == [1.0, -1.0]:
            t_stat_, uncorrected_pvalues = ttest_ind(x, y)
        elif contrast == [-1.0, 1.0]:
            t_stat_, uncorrected_pvalues = ttest_ind(y, x)
        else:
            raise ValueError('Unrecognized contrast !')

        # Fill a dictionnary with the t-statistic and the p-values
        t_test_dictionnary_functional_score[kind] = {'tstatistic': t_stat_, 'uncorrected pvalues': uncorrected_pvalues}

    return t_test_dictionnary_functional_score


def distribution_estimation_mean_subjects_connectivity(mean_matrix_for_each_subjects, groupes, kinds):
    """Provides an estimation of mean and standard deviation of the distribution for the two groupes
    under study, of the mean matrix of each subjects.

    Parameters
    ----------
    mean_matrix_for_each_subjects: dict
        A dictionnary containing, at the first level, the groups as keys, and the kinds as values.
        Inside each each kind, a dictionnary containing the subjects identifier as values, and
        the mean functional connectivity of that subject as values.
    kinds: list
        The list of kinds to test.
    groupes: list
        The list of the two groupes in the study.

    Returns
    -------
    output: dict
        A dictionnary with the groupes as keys, and for each group the estimated standard deviation,
        the estimated mean of the distribution, and the array of shape (n_subject, )
        of the mean matrix of each subjects.


    """
    # Gaussian estimation of the distribution of mean connectivity for each subjects
    estimation_dict = dict.fromkeys(groupes)
    for groupe in groupes:
        estimation_dict[groupe] = dict.fromkeys(kinds)
        for kind in kinds:
            # Fetch the mean connectivity for each subject in one numerical array
            mean_connectivity_vec = np.array(
                [mean_matrix_for_each_subjects[groupe][kind][s]
                 for s in mean_matrix_for_each_subjects[groupe][kind].keys()])
            # Estimate the mean and standard deviation of the data with a Gaussian assumption
            # Fill the dictionnary with the parameters estimate for the current group and kind
            estimation_dict[groupe][kind] = {
                'std': functional_connectivity_distribution_estimation(mean_connectivity_vec)[2],
                'mean': functional_connectivity_distribution_estimation(mean_connectivity_vec)[1],
                'mean connectivity': mean_connectivity_vec}

    return estimation_dict


def intra_network_two_samples_t_test(intra_network_connectivity_dictionary, groupes, kinds,
                                     contrast, network_labels_list, alpha=.05,
                                     p_value_correction_method='fdr_bh', assume_equal_var=True,
                                     nan_policy='omit', paired=False):
    """Test the difference of intra network connectivity between the groups under the study.

    Parameters
    ----------
    intra_network_connectivity_dictionary: dict
        A subjects connectivity dictionnary, with groupes in the study as the first levels of keys,
        the kinds in the study as the second levels of keys, and the intra network connectivity for each
        network in the study.
    groupes: list
        The list of the groups under the study.
    kinds: list
        The list of kinds in the study.
    contrast: list
        The contrast vector use for the t-test. Choices are [1.0, -1.0], or [-1.0, 1.0].
    network_labels_list: list
        The list of the name of the different network.
    alpha: float, optional
        The type I error rate threshold. For p-value under alpha, the null hypothesis can be rejected.
        Default os 0.05.
    p_value_correction_method: string, optional
        The correction method accounting for the multiple comparison problem.
        Default is 'fdr_bh', the traditional False Discovery Rate correction from Benjamini & Hochberg.
    assume_equal_var: bool, optional
        If False, the Welch t-test is perform accounting for different variances between the tested sample.
    nan_policy: string, optional
        Behavior regarding possible missing data (nan values). Default is 'omit'.

    Returns
    -------
    output: dict
        A dictionnary structure containing the t-test result for each network : the t statistic value, the corrected and
        uncorrected p values, the intra network connectivity array of shape (number of subject, ) for
        each group, and the contrast vector.
        .
    # TODO: add an argument to precise the field name in the dictionnary containing the network strength.
    """
    intra_network_strength_t_test = dict.fromkeys(kinds)
    # For some network, like Auditory network in subject 40 in patients, the strength is a masked value which will we
    # convert to nan. So we force to omit nan value in the t-test function call.
    for kind in kinds:

        intra_network_strength_t_test[kind] = dict.fromkeys(network_labels_list)
        # List to stack the uncorrected p value for each network
        all_network_uncorrected_p_values = []
        for network in network_labels_list:
            # Fetch the intra-network strength for each subject in the first group
            x = np.array([intra_network_connectivity_dictionary[groupes[0]][subject][kind][network]['network connectivity strength']
                          for subject in intra_network_connectivity_dictionary[groupes[0]].keys()])
            # Fetch the intra-network strength for each subject in the second group
            y = np.array([intra_network_connectivity_dictionary[groupes[1]][subject][kind][network]['network connectivity strength']
                          for subject in intra_network_connectivity_dictionary[groupes[1]].keys()])
            # Perform two sample t-test according to contrast vector
            if contrast == [1.0, -1.0]:
                if paired is False:
                    network_t_statistic, network_p_values_uncorrected = ttest_ind(x, y,
                                                                                  nan_policy=nan_policy,
                                                                                  equal_var=assume_equal_var)
                else:
                    network_t_statistic, network_p_values_uncorrected = ttest_rel(x, y,
                                                                                  nan_policy=nan_policy,
                                                                                  )
            elif contrast == [-1.0, 1.0]:
                if paired is False:
                    network_t_statistic, network_p_values_uncorrected = ttest_ind(y, x,
                                                                                  nan_policy=nan_policy,
                                                                                  equal_var=assume_equal_var)
                else:
                    network_t_statistic, network_p_values_uncorrected = ttest_rel(y, x,
                                                                                  nan_policy=nan_policy,
                                                                                  )
            else:
                raise ValueError('Unrecognized contrast !')

            # Fill the dictionary to save the result
            intra_network_strength_t_test[kind][network] = {'t statistic': network_t_statistic,
                                                            'uncorrected p values': network_p_values_uncorrected,
                                                            groupes[0]: x,
                                                            groupes[1]: y,
                                                            'contrast': contrast,
                                                            }
            # Stack the uncorrected network p value in the order of network_labels_list
            all_network_uncorrected_p_values.append(network_p_values_uncorrected)

        # Correction of p-values, for each kinds correct for the number of test i.e the number of network
        all_network_p_values = np.array(all_network_uncorrected_p_values)
        # Correct the p value with the chosen method
        reject_boolean_mask, corrected_pvalues, _, _ = \
            multipletests(pvals=all_network_p_values,
                          method=p_value_correction_method,
                          alpha=alpha)

        # Fill the dictionnary, appending a new key containing the corrected p values
        for network in network_labels_list:
            intra_network_strength_t_test[kind][network]['corrected pvalues'] = \
                corrected_pvalues[network_labels_list.index(network)]

    return intra_network_strength_t_test


def inter_network_two_sample_t_test(subjects_inter_network_connectivity_matrices, groupes,
                                    kinds, contrast,
                                    network_label_list, alpha=.05,
                                    p_value_correction_method='fdr_bh',
                                    assuming_equal_var=True,
                                    nan_policy='omit'):
    """Test the difference of connectivity between network performing a two sample t test.

    Parameters
    ----------
    subjects_inter_network_connectivity_matrices: dict
        A subjects connectivity dictionnary, with groupes in the study as the first levels of keys,
        the kinds in the study as the second levels of keys, and the inter network connectivity
        matrices as values, of shape (number of networks, number of networks)
    groupes: list
        The list of the groups under the study.
    kinds: list
        The list of kinds in the study.
    contrast: list
        The contrast vector use for the t-test. Choices are [1.0, -1.0], or [-1.0, 1.0].
    network_label_list: list
        The list of the name of the different network.
    alpha: float, optional
        The type I error rate threshold. For p-value under alpha, the null hypothesis can be rejected.
        Default is 0.05.
    p_value_correction_method: string, optional
        The correction method accounting for the multiple comparison problem.
        Default is 'fdr_bh', the traditional False Discovery Rate correction from
        Benjamini & Hochberg.
    assuming_equal_var: bool, optional
        If False, the Welch t-test is perform accounting for different variances between
        the tested sample.
    nan_policy: string, optional
        Behavior regarding possible missing data (nan values). Default is 'omit'.

    Returns
    -------
    output: dict
        A dictionnary structure containing the t-test result : the raw t statistic array, the corrected and
        uncorrected p values array, the masked t statistic array for corrected p values under
        the alpha threshold.

    """
    inter_network_t_test_result = dict.fromkeys(kinds)

    n_network = len(network_label_list)

    # Fetch the group connectivity, and compute a t-test for each kind.
    for kind in kinds:
        # Fetch and vectorize the subjects connectivity matrices
        x = np.array([vectorizer(numpy_array=subjects_inter_network_connectivity_matrices[groupes[0]][s][kind],
                                 discard_diagonal=True)[1]
                      for s in subjects_inter_network_connectivity_matrices[groupes[0]].keys()])

        y = np.array([vectorizer(numpy_array=subjects_inter_network_connectivity_matrices[groupes[1]][s][kind],
                                 discard_diagonal=True)[1]
                      for s in subjects_inter_network_connectivity_matrices[groupes[1]].keys()])
        if contrast == [1.0, -1.0]:
            t_statistic, p_values_uncorrected = ttest_ind(x, y, axis=0, equal_var=assuming_equal_var,
                                                          nan_policy=nan_policy)
        elif contrast == [-1.0, 1.0]:
            t_statistic, p_values_uncorrected = ttest_ind(y, x, axis=0, equal_var=assuming_equal_var,
                                                          nan_policy=nan_policy)
        else:
            raise ValueError('Unrecognized contrast !')

        # Correction of p values
        vec_reject_mask, vec_corrected_p_values, _, _, = multipletests(pvals=p_values_uncorrected,
                                                                       alpha=alpha,
                                                                       method=p_value_correction_method)

        # Reconstruction of boolean reject_ mask : it's a binary mask
        reject_m_ = vec_to_sym_matrix(vec_reject_mask, np.bool_(np.ones(n_network)))
        # Reconstruction of raw t values
        raw_tvals_m = vec_to_sym_matrix(t_statistic, np.zeros(n_network))
        # Reconstruction of raw p values
        raw_pvals_m = vec_to_sym_matrix(p_values_uncorrected, np.ones(n_network))
        # Reconstruction of corrected p values
        corrected_pvals_m = vec_to_sym_matrix(vec_corrected_p_values, np.ones(n_network))
        # Construction of masked raw t values according to corrected p values under alpha threshold
        significant_tvals_m = np.multiply(raw_tvals_m, reject_m_)

        inter_network_t_test_result[kind] = {'raw t statistic': raw_tvals_m,
                                             'uncorrected p values': raw_pvals_m,
                                             'corrected p values': corrected_pvals_m,
                                             'significant t values': significant_tvals_m,
                                             'contrast': contrast}

    return inter_network_t_test_result


def two_sample_t_test_(connectivity_dictionnary_, groupes, kinds, field, contrast,
                       assume_equal_var=True,
                       nan_policy='omit',
                       paired=False):
    """Perform a simple two sample t test between two sets
    of connectivity matrices.

    Parameters
    ---------
    connectivity_dictionnary_: dict
        A dictionnary which contain some connectivity of interest, and associated measure,
        in a ConPagnon subjects connectivity matrices like.
    groupes: list
        The list of groupes under in the study.
    kinds: list
        The list of kinds in the study
    field: string
        The field name in the dictionary containing the connectivity coefficient array.
        It should be a 1D vector of shape (number of subject, ).
    contrast: list
        The contrast vector. Choices are [1.0, -1.0], or [-1.0, 1.0].
    assume_equal_var: bool, optional
        If False, a Welch t-test assuming different variances between the two group.
    nan_policy: string, optional
        Behavior regarding the missing value in the tested data. Default is 'omit'

    Returns
    -------
    output: dict
        A dictionnary with the raw t statistic, the used contrast vector,
        and the **uncorrected** p values.

    """

    t_test_result = dict.fromkeys(kinds)
    # Fetch the group connectivity, and compute a t-test for each kind.
    for kind in kinds:
        x = connectivity_dictionnary_[groupes[0]][kind][field]
        y = connectivity_dictionnary_[groupes[1]][kind][field]
        if contrast == [1.0, -1.0]:
            if paired is False:

                t_statistic, p_values_uncorrected = ttest_ind(x, y, equal_var=assume_equal_var,
                                                              nan_policy=nan_policy)
            else:
                t_statistic, p_values_uncorrected = ttest_rel(x, y, nan_policy=nan_policy)
        elif contrast == [-1.0, 1.0]:
            if paired is False:

                t_statistic, p_values_uncorrected = ttest_ind(y, x, equal_var=assume_equal_var,
                                                              nan_policy=nan_policy)
            else:
                t_statistic, p_values_uncorrected = ttest_rel(y, x, nan_policy=nan_policy)
        else:
            raise ValueError('Unrecognized contrast !')
        t_test_result[kind] = {'t_statistic': t_statistic,
                               'uncorrected p value': p_values_uncorrected,
                               'contrast': contrast}

    return t_test_result


def regress_confounds(vectorize_subjects_connectivity, confound_dictionary, groupes, kinds,
                      data,
                      sheetname,
                      NA_action='drop'):
    # TODO: add a way to select column containing the subjects ID, and
    # TODO: shift it to be the index of the DataFrame
    """Regress confound on connectivity matrices

    Parameters
    ----------
    vectorize_subjects_connectivity: dict
        The subject connectivity matrices dictionary, with vectorized matrices,
        WITHOUT the diagonal.
    confound_dictionary: dict
        The nested dictionary containing for each group and kind: a field
        named 'confounds' containing a list of confounds. A second field named
        'subjects to drop' containing a list of subjects identifier to drop, None
        if you want to pick all of the subjects.
    groupes: list
        The list of group on which you want to regress confounds.
    kinds: list
        The list of kind
    data: str
        The full path, including extension to the excel file containing
        the confound for each subjects. This will be read by pandas.
        Note that the index of the resulting dataframe must be the
        subjects identifiers.
    sheetname: str
        The sheet name if the excel file containing the confound for each
        subjects of each groups.
    NA_action: str, optional
        Behavior regarding the missing values. If 'drop', the entire
        row is deleted from the design matrix.
    """
    # Reading the data excel file
    df = pd.read_excel(data, sheet_name=sheetname)

    # Copy of the vectorized connectivity matrices dictionary
    # to override the matrices
    regressed_subjects_connectivity_dictionary = copy.deepcopy(vectorize_subjects_connectivity)
    regression_by_kind = dict.fromkeys(groupes, dict.fromkeys(kinds))
    # Save the design matrix, and the stack of matrices before and after
    # regressing
    regression_data_dictionary = dict.fromkeys(groupes)

    for group in groupes:
        # Read the confound and subject to drop for
        # each group
        # List of confounds
        group_confounds = confound_dictionary[group]['confounds']
        # List of subject to drop
        group_subjects_to_drop = confound_dictionary[group]['subjects to drop']
        # List of subjects present in the group dictionary
        group_subjects_list = list(vectorize_subjects_connectivity[group].keys())

        # Copy of original dataframe to avoid
        # side effect :
        df_c = df.copy()
        # Fetch the sub-dataframe containing the subjects
        # in the dictionary
        df_group = df_c.loc[group_subjects_list]
        if group_subjects_to_drop is not None:
            df_group = df_group.drop(labels=group_subjects_to_drop)
        else:
            df_group = df_group

        # Build the formula: regress the confound without
        # Intercept term
        group_formula = '+'.join(group_confounds)

        # Build the design matrix according the dataframe and regression model.
        X_df = dmatrix(formula_like=group_formula, data=df_group,
                       return_type='dataframe',
                       NA_action=NA_action)
        # Drop intercept
        X_df = X_df.drop('Intercept', axis=1)

        # Conversion of design dataframe into numpy array
        # for QR decomposition
        X = np.array(X_df)

        regression_data_dictionary[group] = {'design matrix': X_df}

        for kind in kinds:
            # stack the vectorized array into a numpy array of shape (n_subjects, n_features)
            y = np.array([vectorize_subjects_connectivity[group][subject][kind]
                          for subject in group_subjects_list])

            # check first dimension of X and y, it should be the same
            if y.shape[0] != X.shape[0]:
                raise ValueError('Matrices are not aligned ! First '
                                 'dimension of y is {} and first dimension of X is {}'.format(y.shape[0],
                                                                                              X.shape[0]))

            # QR decomposition of X matrix
            Q, R, _ = linalg.qr(X, mode='economic', pivoting=True)
            # Improve numerical stability
            Q = Q[:, np.abs(np.diag(R)) > np.finfo(np.float).eps * 100.]

            # Regress the confounds in X on y
            y_regressed = y - Q.dot(Q.T).dot(y)

            # Override the connectivity matrices for the current group and kind
            for subject in group_subjects_list:
                regressed_subjects_connectivity_dictionary[group][subject][kind] = \
                    y_regressed[group_subjects_list.index(subject), ...]

            regression_by_kind[group][kind] = {'original matrices': y, 'after regression': y_regressed}

        regression_data_dictionary[group].update(regression_by_kind)

    return regressed_subjects_connectivity_dictionary, regression_data_dictionary


def design_matrix_builder(dataframe, formula, return_type='dataframe'):
    """Build a design matrix based on a dataframe

    Parameters
    ----------
    dataframe: pandas.DataFrame
        A pandas dataframe containing the data.
    formula: str
        The formula written in R style, with the variable
        to explain in the left side of the ~, and the explanatory
        variables in the right side.
    return_type: str, optional
        Return type of the response variable, and design matrix.
        Default is dataframe. Other choices are: matrix.

    Returns
    -------
    output1: pandas.DataFrame
        A dataframe, of shape (n_observations, ). This is
        the variable to explain
    output2: pandas.DataFrame
        The design matrix, of shape (n_observation, n_explanatory_variables +1)
    """
    # Build the design matrix and return the response variable
    # and design matrix
    y, X = dmatrices(formula, data=dataframe, return_type=return_type)

    return y, X


def ols_regression(y, X):
    """Fit a linear model with ordinary least square regression
    from statmodels library

    Parameters
    ----------
    y: array-like
        The variable to explain of shape (n_observations, )
    X: array-like
        The design matrix, of shape (n_obervations, n_regressors) or
        (n_obervations + 1, n_regressors) if intercept is added in the model.

    Returns
    -------
    output: statsmodels.regression.linear_model.RegressionResultsWrapper
        A statsmodels regression object containing the fit of the model.
    """

    # Model initialization with OLS method
    ols_model = sm.OLS(endog=y, exog=X)
    # Fit of the model
    ols_model_fit = ols_model.fit()

    return ols_model_fit


def ols_regression_formula(formula, data):
    """Fit a linear model with a formula API in R style.

    :param formula:
    :param data:
    :return:
    """
    ols_model_fit = smf.ols(formula=formula, data=data).fit()

    return ols_model_fit


def partial_corr(C):
    """
    Returns the sample linear partial correlation coefficients between pairs of variables in C, controlling
    for the remaining variables in C.
    Parameters
    ----------
    C : array-like, shape (n, p)
        Array with the different variables. Each column of C is taken as a variable
    Returns
    -------
    P : array-like, shape (p, p)
        P[i, j] contains the partial correlation of C[:, i] and C[:, j] controlling
        for the remaining variables in C.
    """

    C = np.asarray(C)
    # Add a columns of one to have the same behavior of matlab partialcorr function
    C = np.c_[C, np.ones(C.shape[0])]
    p = C.shape[1]
    P_corr = np.zeros((p, p), dtype=np.float)
    for i in range(p):
        P_corr[i, i] = 1
        for j in range(i + 1, p):
            idx = np.ones(p, dtype=np.bool)
            idx[i] = False
            idx[j] = False
            beta_i = linalg.lstsq(C[:, idx], C[:, j])[0]
            beta_j = linalg.lstsq(C[:, idx], C[:, i])[0]

            res_j = C[:, j] - C[:, idx].dot(beta_i)
            res_i = C[:, i] - C[:, idx].dot(beta_j)

            corr, _ = pearsonr(res_i, res_j)
            P_corr[i, j] = corr
            P_corr[j, i] = corr

    return P_corr



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 16 09:13:14 2013

@author: edouard
"""
import unittest

import numpy as np
from numpy.testing import assert_almost_equal
import mulm
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smfrmla


class TestMULMOLS(unittest.TestCase):

    def test_ttest(self):
        n = 100
        px = 5
        py_info = 2
        py_noize = 100
        beta = np.array([1, 0, -.5] + [0] * (px - 4) + [2]).reshape((px, 1))
        X = np.hstack([np.random.randn(n, px-1), np.ones((n, 1))]) # X with intercept
        Y = np.random.randn(n, py_info + py_noize)
        # Causal model: add X on the first py_info variable
        Y[:, :py_info] += np.dot(X, beta)
        # Two-tailed t-test all the regressors
        contrasts = np.identity(X.shape[1])

        ## OLS with statmodels, need to iterate over Y columns
        sm_tvals = list()
        sm_pvals = list()
        for j in range(Y.shape[1]):
            mod = sm.OLS(Y[:, j], X)
            sm_ttest = mod.fit().t_test(contrasts)
            sm_tvals.append(sm_ttest.tvalue)
            sm_pvals.append(sm_ttest.pvalue)
        sm_tvals = np.asarray(sm_tvals).T
        sm_pvals = np.asarray(sm_pvals).T
        ## OLS with MULM two-tailed
        mod = mulm.MUOLS(Y, X).fit()
        mulm_tvals, mulm_pvals, mulm_df = mod.t_test(contrasts, pval=True, two_tailed=True)

        mod_block = mulm.MUOLS(Y, X).fit(block=True, max_elements=1000)
        mulm_tvals_block, mulm_pvals_block, mulm_df_block = mod_block.t_test(contrasts, pval=True, two_tailed=True)

        # Check that results are similar
        assert_almost_equal(mulm_tvals, sm_tvals)
        assert_almost_equal(mulm_pvals, sm_pvals)

        assert np.all(mulm_tvals == mulm_tvals_block)
        assert np.all(mulm_pvals == mulm_pvals_block)
        assert np.all(mulm_df == mulm_df_block)

#        ## OLS with MULM one-tailed
#        mod = mulm.MUOLS(Y, X).fit()
#        mulm_tvals, mulm_pvals, mulm_df = mod.t_test(contrasts, pval=True, two_tailed=False)
#
#        # Check that results ar similar
#        assert_almost_equal(mulm_tvals, sm_tvals)
#        sm_pvals /= 2
#        #sm_pvals[sm_pvals > 1] = 1
#        assert_almost_equal(mulm_pvals, sm_pvals)
#        mulm_pvals -  np.min(sm_pvals*2,1)


    def test_maxT(self):
        n = 100
        px = 5
        py_info = 2
        py_noize = 100

        beta = np.array([1, 0, -.5] + [0] * (px - 4) + [2]).reshape((px, 1))
        np.random.seed(42)
        X = np.hstack([np.random.randn(n, px-1), np.ones((n, 1))])
        Y = np.random.randn(n, py_info + py_noize)
        # Causal model: add X on the first py_info variable
        Y[:, :py_info] += np.dot(X, beta)
        contrasts = np.identity(X.shape[1])

        mod = mulm.MUOLS(Y, X).fit()
        tvals, rawp, df = mod.t_test(contrasts, pval=True, two_tailed=True)
        tvals2, maxT, df2 = mod.t_test_maxT(contrasts, two_tailed=True)
        assert np.all(tvals == tvals2)
        assert np.all(df == df2)

        mod_block = mulm.MUOLS(Y, X).fit(block=True, max_elements=1000)
        tvals_block, rawp_block, df_block = mod.t_test(contrasts, pval=True, two_tailed=True)
        tvals_block2, maxT_block, df_block2 = mod_block.t_test_maxT(contrasts, two_tailed=True)
        assert np.all(tvals_block == tvals_block2)
        assert np.all(tvals_block2 == tvals2)
        assert np.all(df_block == df_block2)
        assert np.all(df_block2 == df2)

        # More than 10 positive with uncorrected pval
        expected_tp = py_info * 3
        expected_fp = ((py_info + py_noize) * 5 - expected_tp) * 0.05
        expected_p = expected_tp + expected_fp
        # Test the number of rawp positive lie within a expected positive +-10
        assert (np.sum(rawp < 0.05) < (expected_p + 10)) and (np.sum(rawp < 0.05) > (expected_p - 10))
        assert (np.sum(rawp_block < 0.05) < (expected_p + 10)) and (np.sum(rawp_block < 0.05) > (expected_p - 10))

        # Test the number maxT positive lie within a expected true positive +-2
        assert np.sum(maxT < 0.05) < (expected_tp + 2) and np.sum(maxT < 0.05) > (expected_tp - 2)
        assert np.sum(maxT_block < 0.05) < (expected_tp + 2) and np.sum(maxT_block < 0.05) > (expected_tp - 2)

    def test_ttest_ftest_vs_statsmodels(self):
        url = 'https://raw.github.com/neurospin/pystatsml/master/datasets/salary_table.csv'
        df = pd.read_csv(url)

        # Fit with statmodel
        oneway = smfrmla.ols('salary ~ experience + education + management', df).fit()
        #print(oneway.summary())
        aov = sm.stats.anova_lm(oneway, typ=2) # Type 2 ANOVA DataFrame

        # Fit with MULM
        X_df = pd.get_dummies(df.iloc[:, 1:])
        X = np.asarray(X_df).astype(float)
        Y = np.asarray(df.salary)[:, None].astype(float)

        con_exp = np.zeros((X.shape[1], X.shape[1]))
        con_exp[0, 0] = 1

        con_edu = np.zeros((X.shape[1], X.shape[1]))
        con_edu[[1, 2, 3], [1, 2, 3]] = 1

        con_man = np.zeros((X.shape[1], X.shape[1]))
        con_man[[4, 5], [4, 5]] = 1

        import mulm
        mod = mulm.MUOLS(Y, X).fit()
        tvals_exp, rawp_expt, df = mod.t_test([1, 0, 0, 0, 0, 0], pval=True, two_tailed=True)
        fvals_exp, rawp_exp = mod.f_test(con_exp, pval=True)
        fvals_edu, rawp_edu = mod.f_test(con_edu, pval=True)
        fvals_man, rawp_man = mod.f_test(con_man, pval=True)

        assert np.allclose(aov.loc['experience', 'F'], tvals_exp[0] ** 2)
        assert np.allclose(aov.loc['experience', 'PR(>F)'], rawp_expt[0] ** 2)

        assert np.allclose(aov.loc['experience', 'F'], fvals_exp[0])
        assert np.allclose(aov.loc['experience', 'PR(>F)'], rawp_exp[0])

        assert np.allclose(aov.loc['education', 'F'], fvals_edu[0])
        assert np.allclose(aov.loc['education', 'PR(>F)'], rawp_edu[0])

        assert np.allclose(aov.loc['management', 'F'], fvals_man[0])
        assert np.allclose(aov.loc['management', 'PR(>F)'], rawp_man[0])

if __name__ == '__main__':

    unittest.main()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 18 16:38:48 2017

@author: db242421


Le module data_handling fournit des fonctions commode pour stocker les differentes informations relatives au fichiers
utiles (irmf, atlas individuels...) dans une structure facilement utilisable de dictionnaire.

"""




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  6 13:40:49 2017

@author: db242421 (dhaif.bekha@cea.fr)
"""

import numpy as np
from sklearn.model_selection import cross_val_score, StratifiedShuffleSplit
from sklearn.svm import LinearSVC
import matplotlib.pylab as plt
import seaborn as sns


def two_groups_classification(pooled_groups_connectivity_matrices, kinds, labels, n_splits, test_size,train_size, C=1.0,
                              dual=True, fit_intercept=True, loss='squared_hinge', max_iter=1000,
                              penalty='l2', random_state=0, scoring='accuracy'):
    """Perform binary classification on two classes using connectivity coefficient as features.


    Perform classification  on two labels classes using a support vector machine with a linear kernel function.

    Parameters
    ----------
    pooled_groups_connectivity_matrices : dict
         A multi-levels dictionnary organised as follow :
             - The first key is the different groupes to compare.
             - The second key is the different kinds you want to test
             for classification. The values are simply the stacked connectivity
             matrices as a ndarray of shape (..., number of regions, number of regions)
    kinds : list
        The different kind you want to test for classification.
    labels : numpy.array of shape (number of subjects)
        A one-dimensional numpy array of binary labels like 0 for the
        first class and one for the second class.
    n_splits : int
        The number of splitting operations performed by the cross validation
        scheme, StratifiedShuffleSplit.
    test_size : float
        Between 0 and 1, the proportion of the testing set. This is the
        complementary of the train_size : test_size = 1 - train_size.
    train_size : float
        Between 0 and 1, the proportion of the training set. This is the
        complementary of the test_size : train_size = 1 - test_size.
    C : float, optional
        Penalty parameter of the error term.
        Default is 1.0.
    dual : bool, optional
        Algoritm to solve the dual or primal formulation of the problems.
        When n_samples < n_features, prefer dual = True, and False otherwise.
    fit_intercept : bool, optional
        Fit an intercept in the model. If false, the data assumed to be centered.
        Default is True

    loss : str, optional
        The loss function you want to use. Choices are : 'hinge' or 'squared_hinge'.
        Default is 'squared_hinge'.
    max_iter : int, optional
        The maximum iteration you want to run
        Default is 1000.
    penalty : str, optional
        The type of penalty term you want to add in the model. Choices are
        the L1 norm 'l1' or L2 norm 'l2'.
        Default is 'l2'.
    scoring : str, optional
        The maners you want to evaluate the classification by computing
        a score.
        Default is 'accuracy'. See Notes.


    Returns
    -------
    output 1 : list
        List of mean scores for each kind.

    output 2 : dict
        Dictionnary of mean scores with each kind as keys.


    Raises
    ------
    ValueError : If the numbers of the different labels is less than two,
    raises a ValueError. Binary classification requires at least two different classes.

    Notes
    -----
    The classifier, and the Cross validation scheme are from the scikit learn
    library. I encourage the users to consult the scikit learn site cited
    in the references to further details concerning the way using support
    vector machines as classifiers, and the different ways of evaluating classification
    algorithm.

    References
    ----------
    The Scikit learn official documentation.
    [1] http://scikit-learn.org/stable/index.html


    """

    # Dictionary for saving the mean accuracy of classification
    mean_score_dict = {}

    # Check that they are two group to classify, i.e two different labels
    n_labels = len(np.unique(labels))
    if n_labels < 2:
        raise ValueError('At least classification require two classes, only {} label where found !'.format(n_labels))

    # liste qui va contenir la moyenne des scores de classification pour chaque kinds
    mean_scores = []
    # Si les labels sont au format de liste on les convertis en tableaux
    if isinstance(labels, str):
        labels = np.array(labels)

    # K-Fold cross validation schéma:
    cv = StratifiedShuffleSplit(n_splits=n_splits, random_state=random_state, train_size=train_size, test_size=test_size)

    # Boucle pour classer les deux groupes
    for kind in kinds:
        # Initialisation du classifieur:
        svc = LinearSVC(random_state=random_state, penalty=penalty, loss=loss, fit_intercept=fit_intercept, C=C,
                        dual = dual, max_iter = max_iter)
        # On estime les parametres du model et on calcules les scores pour chaque fold par cross-validation
        cv_scores = cross_val_score(estimator = svc, X = pooled_groups_connectivity_matrices[kind], y = labels, cv = cv, scoring = scoring)
        # On calcule la moyenne selon les k-fold et on empile dans mean_scores:
        mean_scores.append(cv_scores.mean())
        mean_score_dict[kind] = cv_scores.mean()

    return mean_scores, mean_score_dict



# -*- coding: utf-8 -*-
"""
Created on Thu May 15 11:35:27 2014

@author:  edouard.duchesnay@cea.fr
@license: BSD-3-Clause
"""
# -*- coding: utf-8 -*-
"""
Created on Thu May 15 11:35:27 2014

@author:  edouard.duchesnay@cea.fr
@license: BSD-3-Clause
"""
import numpy as np
import mulm
import pylab as plt

n = 100
px = 5
py_info = 2
py_noize = 100

beta = np.array([1, 0, .5] + [0] * (px - 4) + [2]).reshape((px, 1))
X = np.hstack([np.random.randn(n, px-1), np.ones((n, 1))]) # X with intercept
Y = np.random.randn(n, py_info + py_noize)
# Causal model: add X on the first py_info variable
Y[:, :py_info] += np.dot(X, beta)

# t-test all the regressors (by default mulm and sm do two-tailed tests)
contrasts = np.identity(X.shape[1])

mod = mulm.MUOLS(Y, X)
tvals, rawp, df = mod.fit().t_test(contrasts, pval=True, two_tailed=True)
tvals, maxTpval, df2 = mod.t_test_maxT(contrasts, two_tailed=True)

n, bins, patches = plt.hist([rawp[0,:], maxTpval[0,:]],
                            color=['blue', 'red'],
                            label=['rawp','maxTpval'])
plt.legend()
plt.show()